# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Palettes for dark and light themes used in Spyder.
"""

# Third-party imports
from qdarkstyle.colorsystem import Blue, Gray
from qdarkstyle.dark.palette import DarkPalette
from qdarkstyle.light.palette import LightPalette

# Local imports
from spyder.config.gui import is_dark_interface
from spyder.utils.color_system import (Green, Red, Orange, GroupDark,
                                       GroupLight, Logos)

# =============================================================================
# ---- Spyder palettes
# =============================================================================
class SpyderPaletteDark(DarkPalette):
    """Dark palette for Spyder."""

    # Colors for information and feedback in dialogs
    COLOR_SUCCESS_1 = Green.B40
    COLOR_SUCCESS_2 = Green.B70
    COLOR_SUCCESS_3 = Green.B90

    COLOR_ERROR_1 = Red.B40
    COLOR_ERROR_2 = Red.B70
    COLOR_ERROR_3 = Red.B110

    COLOR_WARN_1 = Orange.B40
    COLOR_WARN_2 = Orange.B70
    COLOR_WARN_3 = Orange.B90
    COLOR_WARN_4 = Orange.B100

    # Icon colors
    ICON_1 = Gray.B140
    ICON_2 = Blue.B80
    ICON_3 = Green.B80
    ICON_4 = Red.B70
    ICON_5 = Orange.B70
    ICON_6 = Gray.B30
    ICON_7 = GroupDark.B90

    # Colors for icons and variable explorer in dark mode
    GROUP_1 = GroupDark.B10
    GROUP_2 = GroupDark.B20
    GROUP_3 = GroupDark.B30
    GROUP_4 = GroupDark.B40
    GROUP_5 = GroupDark.B50
    GROUP_6 = GroupDark.B60
    GROUP_7 = GroupDark.B70
    GROUP_8 = GroupDark.B80
    GROUP_9 = GroupDark.B90
    GROUP_10 = GroupDark.B100
    GROUP_11 = GroupDark.B110
    GROUP_12 = GroupDark.B120

    # Colors for highlight in editor
    COLOR_HIGHLIGHT_1 = Blue.B10
    COLOR_HIGHLIGHT_2 = Blue.B20
    COLOR_HIGHLIGHT_3 = Blue.B30
    COLOR_HIGHLIGHT_4 = Blue.B50

    # Colors for occurrences from find widget
    COLOR_OCCURRENCE_1 = Gray.B10
    COLOR_OCCURRENCE_2 = Gray.B20
    COLOR_OCCURRENCE_3 = Gray.B30
    COLOR_OCCURRENCE_4 = Gray.B50
    COLOR_OCCURRENCE_5 = Gray.B80

    # Colors for Spyder and Python logos
    PYTHON_LOGO_UP = Logos.B10
    PYTHON_LOGO_DOWN = Logos.B20
    SPYDER_LOGO_BACKGROUND = Logos.B30
    SPYDER_LOGO_WEB = Logos.B40
    SPYDER_LOGO_SNAKE = Logos.B50

    # For special tabs
    SPECIAL_TABS_SEPARATOR = Gray.B70
    SPECIAL_TABS_SELECTED = DarkPalette.COLOR_ACCENT_2

    # For the heart used to ask for donations
    COLOR_HEART = Blue.B80

    # For editor tooltips
    TIP_TITLE_COLOR = Green.B80
    TIP_CHAR_HIGHLIGHT_COLOR = Orange.B90


class SpyderPaletteLight(LightPalette):
    """Light palette for Spyder."""

    # Colors for information and feedback in dialogs
    COLOR_SUCCESS_1 = Green.B40
    COLOR_SUCCESS_2 = Green.B70
    COLOR_SUCCESS_3 = Green.B30

    COLOR_ERROR_1 = Red.B40
    COLOR_ERROR_2 = Red.B70
    COLOR_ERROR_3 = Red.B110

    COLOR_WARN_1 = Orange.B40
    COLOR_WARN_2 = Orange.B70
    COLOR_WARN_3 = Orange.B50
    COLOR_WARN_4 = Orange.B40

    # Icon colors
    ICON_1 = Gray.B30
    ICON_2 = Blue.B50
    ICON_3 = Green.B30
    ICON_4 = Red.B70
    ICON_5 = Orange.B70
    ICON_6 = Gray.B140
    ICON_7 = GroupLight.B90

    # Colors for icons and variable explorer in light mode
    GROUP_1 = GroupLight.B10
    GROUP_2 = GroupLight.B20
    GROUP_3 = GroupLight.B30
    GROUP_4 = GroupLight.B40
    GROUP_5 = GroupLight.B50
    GROUP_6 = GroupLight.B60
    GROUP_7 = GroupLight.B70
    GROUP_8 = GroupLight.B80
    GROUP_9 = GroupLight.B90
    GROUP_10 = GroupLight.B100
    GROUP_11 = GroupLight.B110
    GROUP_12 = GroupLight.B120

    # Colors for highlight in editor
    COLOR_HIGHLIGHT_1 = Blue.B140
    COLOR_HIGHLIGHT_2 = Blue.B130
    COLOR_HIGHLIGHT_3 = Blue.B120
    COLOR_HIGHLIGHT_4 = Blue.B110

    # Colors for occurrences from find widget
    COLOR_OCCURRENCE_1 = Gray.B120
    COLOR_OCCURRENCE_2 = Gray.B110
    COLOR_OCCURRENCE_3 = Gray.B100
    COLOR_OCCURRENCE_4 = Gray.B90
    COLOR_OCCURRENCE_5 = Gray.B60

    # Colors for Spyder and Python logos
    PYTHON_LOGO_UP = Logos.B10
    PYTHON_LOGO_DOWN = Logos.B20
    SPYDER_LOGO_BACKGROUND = Logos.B30
    SPYDER_LOGO_WEB = Logos.B40
    SPYDER_LOGO_SNAKE = Logos.B50

    # For special tabs
    SPECIAL_TABS_SEPARATOR = Gray.B70
    SPECIAL_TABS_SELECTED = LightPalette.COLOR_ACCENT_5

    # For the heart used to ask for donations
    COLOR_HEART = Red.B70

    # For editor tooltips
    TIP_TITLE_COLOR = Green.B20
    TIP_CHAR_HIGHLIGHT_COLOR = Orange.B30


# =============================================================================
# ---- Exported classes
# =============================================================================
if is_dark_interface():
    SpyderPalette = SpyderPaletteDark
else:
    SpyderPalette = SpyderPaletteLight



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Pages for the dialog that handles remote connections."""

# Standard library imports
from __future__ import annotations
from collections.abc import Iterable
import os.path as osp
import re
from typing import TypedDict
import uuid

# Third party imports
import asyncssh
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QButtonGroup,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QStackedWidget,
    QStackedLayout,
    QVBoxLayout,
    QWidget,
)

# Local imports
from spyder.api.fonts import SpyderFontType, SpyderFontsMixin
from spyder.api.translations import _
from spyder.api.utils import get_class_values
from spyder.plugins.remoteclient.api.protocol import (
    ConnectionInfo,
    ConnectionStatus,
    ClientType,
    RemoteClientLog,
)
from spyder.plugins.remoteclient.widgets import AuthenticationMethod
from spyder.plugins.remoteclient.widgets.connectionstatus import (
    ConnectionStatusWidget,
)
from spyder.utils.icon_manager import ima
from spyder.utils.stylesheet import AppStyle, MAC, WIN
from spyder.widgets.config import SpyderConfigPage
from spyder.widgets.helperwidgets import MessageLabel, TipWidget

try:
    from spyder_env_manager.spyder.widgets.new_environment import (
        NewEnvironment,
    )
    from spyder_env_manager.spyder.widgets.edit_environment import (
        EditEnvironment,
    )
    ENV_MANAGER = True
except Exception:
    ENV_MANAGER = False


# =============================================================================
# ---- Constants
# =============================================================================
class ValidationReasons(TypedDict):
    repeated_name: bool | None
    missing_info: bool | None
    invalid_address: bool | None
    invalid_url: bool | None
    file_not_found: bool | None
    keyfile_passphrase_is_wrong: bool | None
    keyfile_not_load: bool | None
    keyfile_malformed: bool | None


class CreateEnvMethods:
    NewEnv = 1
    ImportEnv = 2
    NoEnv = 4


# =============================================================================
# ---- Pages
# =============================================================================
class BaseConnectionPage(SpyderConfigPage, SpyderFontsMixin):
    """Base class to create connection pages."""

    MIN_HEIGHT = 450
    NEW_CONNECTION = False
    CONF_SECTION = "remoteclient"

    def __init__(self, parent, host_id=None):
        super().__init__(parent)

        # host_id is None only for the new connection page
        if host_id is None:
            self.host_id = str(uuid.uuid4())
            self.status = ConnectionStatus.Inactive
            self._client_type = None
        else:
            self.host_id = host_id
            self.status = self.get_option(
                f"{host_id}/status", default=ConnectionStatus.Inactive
            )
            self._client_type = self.get_option(
                f"{host_id}/client_type", default=ClientType.SSH
            )

        self._auth_methods = None
        self._widgets_for_validation: dict[str, list[QWidget]] = {}
        self._validation_labels: dict[str: QWidget] = {}
        self._name_widgets: dict[str, QWidget] = {}
        self._address_widgets: dict[str, QWidget] = {}
        self._port_widgets: dict[str, QWidget] = {}
        self._username_widgets: dict[str, QWidget] = {}
        self._url_widgets: dict[str, QWidget] = {}

    # ---- Public API
    # -------------------------------------------------------------------------
    def auth_method(self, from_gui=False):
        if from_gui:
            if self.get_client_type() == ClientType.SSH or (
                self._auth_methods and self._auth_methods.combobox.isVisible()
            ):
                if self._auth_methods.combobox.currentIndex() == 0:
                    auth_method = AuthenticationMethod.Password
                elif self._auth_methods.combobox.currentIndex() == 1:
                    auth_method = AuthenticationMethod.KeyFile
                else:
                    auth_method = AuthenticationMethod.ConfigFile
            else:
                auth_method = AuthenticationMethod.JupyterHub
        else:
            auth_method = self.get_option(f"{self.host_id}/auth_method")

        return auth_method

    def validate_page(self):
        """Validate contents before saving the connection."""
        # Get widgets we're going to interact with
        auth_method = self.auth_method(from_gui=True)
        widgets = self._widgets_for_validation[auth_method]
        validate_label = self._validation_labels[auth_method]

        # Hide label in case the validation pass
        validate_label.setVisible(False)

        reasons: ValidationReasons = {}
        for widget in widgets:
            if not widget.textbox.text():
                # Validate that the required fields are not empty
                widget.status_action.setVisible(True)
                widget.status_action.setToolTip(_("This field is empty"))
                reasons["missing_info"] = True
            elif widget == self._name_widgets[auth_method]:
                # Validate the server name is different from the ones already
                # introduced
                widget.status_action.setVisible(False)
                current_name = widget.textbox.text()
                if not self._validate_name(current_name):
                    reasons["repeated_name"] = True
                    widget.status_action.setVisible(True)
            elif widget == self._address_widgets.get(auth_method):
                # Validate address
                widget.status_action.setVisible(False)
                address = widget.textbox.text()
                if not self._validate_address(address):
                    reasons["invalid_address"] = True
                    widget.status_action.setVisible(True)
            elif widget == self._url_widgets.get(auth_method):
                # Validate URL
                widget.status_action.setVisible(False)
                url = widget.textbox.text()
                if not self._validate_url(url):
                    reasons["invalid_url"] = True
                    widget.status_action.setVisible(True)
            elif (
                auth_method == AuthenticationMethod.KeyFile
                and widget == self._keyfile
            ):
                # Validate Keyfile
                widget.status_action.setVisible(False)
                key_path = widget.textbox.text()

                if not osp.isfile(key_path):
                    reasons["file_not_found"] = True
                    widget.status_action.setVisible(True)
                else:
                    # Attempt to load the key file
                    passphrase = self._passphrase.textbox.text()
                    try:
                        asyncssh.read_private_key(key_path, passphrase)
                    except asyncssh.KeyEncryptionError:
                        # Passphrase is wrong or missing
                        reasons["keyfile_passphrase_is_wrong"] = True
                        widget.status_action.setVisible(True)
                    except asyncssh.KeyImportError:
                        # File is not a real key file
                        reasons["keyfile_not_load"] = True
                        widget.status_action.setVisible(True)
                    except asyncssh.SFTPBadMessage:
                        # Key file is malformed or not a recognized format
                        reasons["keyfile_malformed"] = True
                        widget.status_action.setVisible(True)
            else:
                widget.status_action.setVisible(False)

        if reasons:
            validate_label.set_text(
                self._compose_failed_validation_text(reasons)
            )
            validate_label.setVisible(True)

            # Adjust page height according to the authentication method and
            # number of reasons.
            n_reasons = list(reasons.values()).count(True)
            min_height = self.MIN_HEIGHT
            if self.get_client_type() == ClientType.SSH:
                if (
                    self.auth_method(from_gui=True)
                    == AuthenticationMethod.Password
                ):
                    if n_reasons > 2:
                        min_height = self.MIN_HEIGHT if (WIN or MAC) else 600
                else:
                    if n_reasons == 1:
                        min_height = 640 if MAC else (580 if WIN else 620)
                    else:
                        min_height = 700 if MAC else (620 if WIN else 680)

                self.setMinimumHeight(min_height)

        return False if reasons else True

    def create_jupyterhub_connection_info_widget(self):
        """
        Create widget that contains all other widgets to receive or display
        JupyterHub server based connection info.
        """
        if self.NEW_CONNECTION:
            # Widgets
            intro_label = QLabel(
                _("Configure settings for connecting to a JupyterHub server")
            )
            intro_tip_text = _(
                "Spyder will use this connection to start remote kernels in "
                "the IPython Console. This allows you to use Spyder locally "
                "while running code remotely via a JupyterHub server."
            )
            intro_tip = TipWidget(
                tip_text=intro_tip_text,
                icon=ima.icon('info_tip'),
                hover_icon=ima.icon('info_tip_hover'),
                size=AppStyle.ConfigPageIconSize + 2,
                wrap_text=True,
            )

            # Increase font size to make it more relevant
            font = self.get_font(SpyderFontType.Interface)
            font.setPointSize(font.pointSize() + 1)
            intro_label.setFont(font)

            # Layout
            intro_layout = QHBoxLayout()
            intro_layout.setContentsMargins(0, 0, 0, 0)
            intro_layout.setSpacing(0)
            intro_layout.setAlignment(Qt.AlignCenter)
            intro_layout.addWidget(intro_label)
            intro_layout.addWidget(intro_tip)

        # Final layout
        layout = QVBoxLayout()
        layout.setContentsMargins(
            3 * AppStyle.MarginSize, 0, 3 * AppStyle.MarginSize, 0
        )
        if self.NEW_CONNECTION:
            layout.addLayout(intro_layout)
            layout.addSpacing(8 * AppStyle.MarginSize)
        layout.addWidget(self._create_jupyterhub_subpage())

        jupyterhub_connection_info_widget = QWidget(self)
        jupyterhub_connection_info_widget.setLayout(layout)

        return jupyterhub_connection_info_widget

    def create_ssh_connection_info_widget(self):
        """
        Create widget that contains all other widgets to receive or display
        SSH based connection info.
        """
        # Show intro text and tip for new connections
        if self.NEW_CONNECTION:
            # Widgets
            intro_label = QLabel(
                _("Configure SSH settings for connecting to remote hosts")
            )
            intro_tip_text = _(
                "Spyder will use this connection to start remote kernels in "
                "the IPython Console. This allows you to use Spyder locally "
                "while running code remotely, such as on a cloud instance, "
                "office workstation or high-performance cluster."
            )
            intro_tip = TipWidget(
                tip_text=intro_tip_text,
                icon=ima.icon('info_tip'),
                hover_icon=ima.icon('info_tip_hover'),
                size=AppStyle.ConfigPageIconSize + 2,
                wrap_text=True,
            )

            # Increase font size to make it more relevant
            font = self.get_font(SpyderFontType.Interface)
            font.setPointSize(font.pointSize() + 1)
            intro_label.setFont(font)

            # Layout
            intro_layout = QHBoxLayout()
            intro_layout.setContentsMargins(0, 0, 0, 0)
            intro_layout.setSpacing(0)
            intro_layout.setAlignment(Qt.AlignCenter)
            intro_layout.addWidget(intro_label)
            intro_layout.addWidget(intro_tip)

        # Authentication methods
        # TODO: The config file method is not implemented yet, so we need to
        # disable it for now.
        methods = (
            (_('Password'), AuthenticationMethod.Password),
            (_('Key file'), AuthenticationMethod.KeyFile),
            # (_('Configuration file'), AuthenticationMethod.ConfigFile),
        )

        self._auth_methods = self.create_combobox(
            _("Authentication method:"),
            methods,
            f"{self.host_id}/auth_method",
            items_elide_mode=Qt.ElideNone,
        )

        # Subpages
        password_subpage = self._create_password_subpage()
        keyfile_subpage = self._create_keyfile_subpage()
        configfile_subpage = self._create_configfile_subpage()

        subpages = QStackedWidget(self)
        subpages.addWidget(password_subpage)
        subpages.addWidget(keyfile_subpage)
        subpages.addWidget(configfile_subpage)

        # Signals
        self._auth_methods.combobox.currentIndexChanged.connect(
            subpages.setCurrentIndex
        )
        self._auth_methods.combobox.currentIndexChanged.connect(
            self._copy_info
        )

        # Show password subpage by default for new connections
        if self.NEW_CONNECTION:
            self._auth_methods.combobox.setCurrentIndex(0)

        # Final layout
        right_margin = 3
        if ENV_MANAGER and self.NEW_CONNECTION:
            right_margin = 4 if MAC else (6 if WIN else 5)

        layout = QVBoxLayout()
        layout.setContentsMargins(
            3 * AppStyle.MarginSize, 0, right_margin * AppStyle.MarginSize, 0
        )
        if self.NEW_CONNECTION:
            layout.addLayout(intro_layout)
            layout.addSpacing(8 * AppStyle.MarginSize)
        layout.addWidget(self._auth_methods)
        layout.addSpacing(5 * AppStyle.MarginSize)
        layout.addWidget(subpages)

        ssh_connection_info_widget = QWidget(self)
        ssh_connection_info_widget.setLayout(layout)

        return ssh_connection_info_widget

    def get_client_type(self):
        if self.NEW_CONNECTION:
            # In this case the client type hasn't been saved to our config
            # system, so we need to get it from the UI.
            client_type = (
                ClientType.SSH
                if self.tabs.currentIndex() == 0
                else ClientType.JupyterHub
            )
        else:
            client_type = self._client_type

        return client_type

    # ---- Private API
    # -------------------------------------------------------------------------
    def _create_common_elements(self, auth_method: str):
        """Common elements for the password and keyfile subpages."""
        # Widgets
        name = self.create_lineedit(
            text=_("Name *"),
            option=f"{self.host_id}/{auth_method}/name",
            tip=_("Introduce a name to identify your connection"),
            validate_callback=self._validate_name,
            validate_reason=_("This connection name is already taken"),
        )

        address = self.create_lineedit(
            text=_("Remote address *"),
            option=f"{self.host_id}/{auth_method}/address",
            tip=_(
                "This is the IP address or domain name of your remote machine"
            ),
            validate_callback=self._validate_address,
            validate_reason=_("The address is not a valid IP or domain name"),
        )

        port = self.create_spinbox(
            prefix=_("Port"),
            suffix="",
            option=f"{self.host_id}/{auth_method}/port",
            min_=1,
            max_=65535
        )
        port.spinbox.setStyleSheet("margin-left: 5px")

        username = self.create_lineedit(
            text=_("Username *"),
            option=f"{self.host_id}/{auth_method}/username",
            status_icon=ima.icon("error"),
        )

        self._widgets_for_validation[f"{auth_method}"] = [
            name,
            address,
            username,
        ]
        self._name_widgets[f"{auth_method}"] = name
        self._address_widgets[f"{auth_method}"] = address
        self._port_widgets[f"{auth_method}"] = port
        self._username_widgets[f"{auth_method}"] = username

        # Set 22 as the default port for new conenctions
        if not self.LOAD_FROM_CONFIG:
            port.spinbox.setValue(22)

        # Hide the container widgets because we only use their components
        address.hide()
        port.hide()

        # Layout for the address label
        address_label_layout = QHBoxLayout()
        address_label_layout.setSpacing(0)
        address_label_layout.addWidget(address.label)
        address_label_layout.addWidget(address.help_label)
        address_label_layout.addStretch()

        # Address layout
        address_layout = QGridLayout()
        address_layout.setContentsMargins(0, 0, 0, 0)

        address_layout.addLayout(address_label_layout, 0, 0)
        address_layout.addWidget(address.textbox, 1, 0)
        address_layout.addWidget(port.plabel, 0, 1)
        address_layout.addWidget(port.spinbox, 1, 1)

        return name, address_layout, username

    def _create_password_subpage(self):
        # Widgets
        name, address_layout, username = self._create_common_elements(
            auth_method=AuthenticationMethod.Password
        )

        password = self.create_lineedit(
            text=_("Password *"),
            option=f"{self.host_id}/password",
            tip=(
                _("Your password will be saved securely by Spyder")
                if self.NEW_CONNECTION
                else _("Your password is saved securely by Spyder")
            ),
            status_icon=ima.icon("error"),
            password=True
        )

        validation_label = MessageLabel(self)

        # Add widgets to their required dicts
        self._widgets_for_validation[AuthenticationMethod.Password].append(
            password
        )

        self._validation_labels[
            AuthenticationMethod.Password
        ] = validation_label

        # Layout
        password_layout = QVBoxLayout()
        password_layout.setContentsMargins(0, 0, 0, 0)

        password_layout.addWidget(name)
        password_layout.addSpacing(5 * AppStyle.MarginSize)
        password_layout.addLayout(address_layout)
        password_layout.addSpacing(5 * AppStyle.MarginSize)
        password_layout.addWidget(username)
        password_layout.addSpacing(5 * AppStyle.MarginSize)
        password_layout.addWidget(password)
        password_layout.addSpacing(7 * AppStyle.MarginSize)
        password_layout.addWidget(validation_label)
        password_layout.addStretch()

        password_subpage = QWidget(self)
        password_subpage.setLayout(password_layout)

        return password_subpage

    def _create_keyfile_subpage(self):
        # Widgets
        name, address_layout, username = self._create_common_elements(
            auth_method=AuthenticationMethod.KeyFile
        )

        self._keyfile = self.create_browsefile(
            text=_("Key file *"),
            option=f"{self.host_id}/keyfile",
            alignment=Qt.Vertical,
            status_icon=ima.icon("error"),
        )

        self._passphrase = self.create_lineedit(
            text=_("Passphrase"),
            option=f"{self.host_id}/passphrase",
            tip=(
                _("Your passphrase will be saved securely by Spyder")
                if self.NEW_CONNECTION
                else _("Your passphrase is saved securely by Spyder")
            ),
            password=True
        )

        validation_label = MessageLabel(self)

        # Add widgets to their required dicts
        self._widgets_for_validation[AuthenticationMethod.KeyFile].append(
            self._keyfile
        )
        self._validation_labels[AuthenticationMethod.KeyFile] = (
            validation_label
        )

        # Layout
        keyfile_layout = QVBoxLayout()
        keyfile_layout.setContentsMargins(0, 0, 0, 0)

        keyfile_layout.addWidget(name)
        keyfile_layout.addSpacing(5 * AppStyle.MarginSize)
        keyfile_layout.addLayout(address_layout)
        keyfile_layout.addSpacing(5 * AppStyle.MarginSize)
        keyfile_layout.addWidget(username)
        keyfile_layout.addSpacing(5 * AppStyle.MarginSize)
        keyfile_layout.addWidget(self._keyfile)
        keyfile_layout.addSpacing(5 * AppStyle.MarginSize)
        keyfile_layout.addWidget(self._passphrase)
        keyfile_layout.addSpacing(7 * AppStyle.MarginSize)
        keyfile_layout.addWidget(validation_label)
        keyfile_layout.addStretch()

        keyfile_subpage = QWidget(self)
        keyfile_subpage.setLayout(keyfile_layout)

        return keyfile_subpage

    def _create_configfile_subpage(self):
        # Widgets
        name = self.create_lineedit(
            text=_("Name *"),
            option=f"{self.host_id}/{AuthenticationMethod.ConfigFile}/name",
            tip=_("Introduce a name to identify your connection"),
            status_icon=ima.icon("error"),
        )

        configfile = self.create_browsefile(
            text=_("Configuration file *"),
            option=f"{self.host_id}/configfile",
            alignment=Qt.Vertical,
            status_icon=ima.icon("error"),
        )

        validation_label = MessageLabel(self)

        # Add widgets to their required dicts
        self._name_widgets[AuthenticationMethod.ConfigFile] = name
        self._widgets_for_validation[AuthenticationMethod.ConfigFile] = [
            name,
            configfile,
        ]
        self._validation_labels[
            AuthenticationMethod.ConfigFile
        ] = validation_label

        # Layout
        configfile_layout = QVBoxLayout()
        configfile_layout.setContentsMargins(0, 0, 0, 0)
        configfile_layout.addWidget(name)
        configfile_layout.addSpacing(5 * AppStyle.MarginSize)
        configfile_layout.addWidget(configfile)
        configfile_layout.addSpacing(7 * AppStyle.MarginSize)
        configfile_layout.addWidget(validation_label)
        configfile_layout.addStretch()

        configfile_widget = QWidget(self)
        configfile_widget.setLayout(configfile_layout)

        return configfile_widget

    def _create_jupyterhub_subpage(self):
        # Widgets
        name = self.create_lineedit(
            text=_("Name *"),
            option=f"{self.host_id}/{AuthenticationMethod.JupyterHub}/name",
            tip=_("Introduce a name to identify your connection"),
            status_icon=ima.icon("error"),
        )

        url = self.create_lineedit(
            text=_("Server URL *"),
            option=f"{self.host_id}/url",
            tip=_("This is the URL of the JupyterHub server"),
            validate_callback=self._validate_url,
            validate_reason=_("The URL is not a valid one"),
        )

        token = self.create_lineedit(
            text=_("Token *"),
            option=f"{self.host_id}/token",
            tip=(
                _("Your token will be saved securely by Spyder")
                if self.NEW_CONNECTION
                else _("Your token is saved securely by Spyder")
            ),
            status_icon=ima.icon("error"),
            password=True
        )

        validation_label = MessageLabel(self)

        # Add widgets to their required dicts
        self._name_widgets[AuthenticationMethod.JupyterHub] = name
        self._widgets_for_validation[AuthenticationMethod.JupyterHub] = [
            name,
            url,
            token,
        ]
        self._url_widgets[f"{AuthenticationMethod.JupyterHub}"] = url
        self._validation_labels[
            AuthenticationMethod.JupyterHub
        ] = validation_label

        # Layout
        jupyterhub_layout = QVBoxLayout()
        jupyterhub_layout.setContentsMargins(0, 0, 0, 0)
        jupyterhub_layout.addWidget(name)
        jupyterhub_layout.addSpacing(5 * AppStyle.MarginSize)
        jupyterhub_layout.addWidget(url)
        jupyterhub_layout.addSpacing(5 * AppStyle.MarginSize)
        jupyterhub_layout.addWidget(token)
        jupyterhub_layout.addSpacing(7 * AppStyle.MarginSize)
        jupyterhub_layout.addWidget(validation_label)
        jupyterhub_layout.addStretch()

        jupyterhub_widget = QWidget(self)
        jupyterhub_widget.setLayout(jupyterhub_layout)

        return jupyterhub_widget

    def _validate_name(self, name):
        """Check connection name is not taken by a previous connection."""
        servers = self.get_option("servers", default={})
        for server in servers:
            # Don't check repeated name with the same server
            if server == self.host_id:
                continue

            names = [
                self.get_option(f"{server}/{method}/name", default="")
                for method in get_class_values(AuthenticationMethod)
            ]

            if name in names:
                return False

        return True

    def _validate_url(self, url):
        # Regex pattern for a valid URL.
        # See https://learn.microsoft.com/en-us/previous-versions/msp-n-p/
        # ff650303(v=pandp.10)#common-regular-expressions
        url_pattern = (
            r'^(ht|f)tp(s?)\:\/\/'
            r'[0-9a-zA-Z]([-.\w]*[0-9a-zA-Z])*(:(0-9)*)*(\/?)'
            r'([a-zA-Z0-9\-\.\?\,\'\/\\\+&amp;%\$#_]*)?$'
        )
        url_re = re.compile(url_pattern)

        return True if url_re.match(url) else False

    def _validate_address(self, address):
        """Validate if address introduced by users is correct."""
        # Regex pattern for a valid domain name (simplified version)
        domain_pattern = (
            r'^([a-zA-Z0-9][a-zA-Z0-9-]{0,61}[a-zA-Z0-9]\.){1,}[a-zA-Z]{2,}$'
        )

        # Regex pattern for a valid IPv4 address
        ipv4_pattern = r'^(\d{1,3}\.){3}\d{1,3}$'

        # Regex pattern for a valid IPv6 address (simplified version)
        ipv6_pattern = r'^([\da-fA-F]{1,4}:){7}[\da-fA-F]{1,4}$'

        # Combined pattern to check all three formats
        combined_pattern = (
            f'({domain_pattern})|({ipv4_pattern})|({ipv6_pattern})'
        )

        address_re = re.compile(combined_pattern)
        return True if address_re.match(address) else False

    def _compose_failed_validation_text(self, reasons: ValidationReasons):
        """
        Compose validation text from a dictionary of reasons for which it
        failed.
        """
        n_reasons = list(reasons.values()).count(True)
        prefix = "- " if n_reasons > 1 else ""
        suffix = "<br>" if n_reasons > 1 else ""

        text = ""
        if reasons.get("repeated_name"):
            text += (
                prefix
                + _(
                    "The name you selected is already used by another "
                    "connection."
                )
                + suffix
            )

        if reasons.get("invalid_address"):
            text += (
                prefix
                + _(
                    "The address you provided is not a valid IP or domain "
                    "name."
                )
                + suffix
            )

        if reasons.get("invalid_url"):
            text += (
                prefix + _("The URL you provided is not valid.") + suffix
            )

        if reasons.get("file_not_found"):
            text += (
                prefix + _("The file you provided is not available.") + suffix
            )

        if reasons.get("keyfile_passphrase_is_wrong"):
            text += (
                prefix
                + _(
                    "The key file is encrypted but no passphrase was provided "
                    "or it's incorrect"
                )
                + suffix
            )

        if reasons.get("keyfile_not_load"):
            text += (
                prefix + _("The key file you provided is invalid.") + suffix
            )

        if reasons.get("keyfile_malformed"):
            text += (
                prefix
                + _(
                    "The key file you provided is malformed or not a valid "
                    "SSH key format."
                )
                + suffix
            )

        # IMPORTANT: This needs to remain as the last reason because it doesn't
        # have a suffix. It's also the least important.
        if reasons.get("missing_info"):
            text += prefix + _("There are missing fields on this page.")

        return text

    def _copy_info(self, auth_method_index: int):
        """
        Copy common info from one authentication method to another.

        Notes
        -----
        * This makes it easier to switch to a different method when creating or
          editing a connection, e.g. if you realized you were entering info in
          the wrong method when creating a new connection.
        """
        # Get current and previous authentication methods
        current_method = (
            AuthenticationMethod.Password
            if auth_method_index == 0
            else AuthenticationMethod.KeyFile
        )
        previous_method = (
            AuthenticationMethod.Password
            if current_method == AuthenticationMethod.KeyFile
            else AuthenticationMethod.KeyFile
        )

        # Copy info from the previous method to the current one.
        previous_name = self._name_widgets[previous_method].textbox.text()
        if previous_name:
            self._name_widgets[current_method].textbox.setText(previous_name)

        previous_address = self._address_widgets[
            previous_method
        ].textbox.text()
        if previous_address:
            self._address_widgets[current_method].textbox.setText(
                previous_address
            )

        previous_port = self._port_widgets[previous_method].spinbox.value()
        if previous_port != 22:
            self._port_widgets[current_method].spinbox.setValue(previous_port)

        previous_username = self._username_widgets[
            previous_method
        ].textbox.text()
        if previous_username:
            self._username_widgets[current_method].textbox.setText(
                previous_username
            )


class NewConnectionPage(BaseConnectionPage):
    """Page to receive SSH credentials for a remote connection."""

    MAX_WIDTH = 600 if MAC else 580
    LOAD_FROM_CONFIG = False
    NEW_CONNECTION = True

    # ---- SidebarPage API
    # -------------------------------------------------------------------------
    def get_name(self):
        return _("New connection")

    def setup_page(self):
        # Attributes
        self.env_method_group = QButtonGroup(self)
        self._radio_buttons_to_info_widgets: dict[
            CreateEnvMethods, QWidget
        ] = {}

        # Widgets
        self.ssh_info_widget = self.create_ssh_connection_info_widget()
        jupyterhub_info_widget = self.create_jupyterhub_connection_info_widget()

        if ENV_MANAGER:
            self.env_creation_widget = self._create_env_creation_widget()
            self.env_packages_widget = self._create_env_packages_widget()

        # Use a stacked widget/layout so we can hide the current widgets and
        # create new ones in case users want to introduce more connections.
        self.ssh_widget = QStackedWidget(self)
        self.ssh_widget.addWidget(self.ssh_info_widget)
        if ENV_MANAGER:
            self.ssh_widget.addWidget(self.env_creation_widget)
            self.ssh_widget.addWidget(self.env_packages_widget)

        self.jupyterhub_widget = QWidget(self)
        jupyterhub_layout = QStackedLayout()
        jupyterhub_layout.addWidget(jupyterhub_info_widget)
        self.jupyterhub_widget.setLayout(jupyterhub_layout)

        self.create_tab("SSH", self.ssh_widget)
        self.create_tab("JupyterHub", self.jupyterhub_widget)

    def get_icon(self):
        return self.create_icon("add_server")

    # ---- SpyderConfigPage API
    # -------------------------------------------------------------------------
    def save_to_conf(self):
        super().save_to_conf()

        if self.NEW_CONNECTION:
            # Set the client type for new connections following current tab
            # index
            client_type = self.get_client_type()
            self.set_option(f"{self.host_id}/client_type", client_type)
            if client_type == ClientType.JupyterHub:
                # Set correct auth_method option following client type detected
                self.set_option(
                    f"{self.host_id}/auth_method",
                    AuthenticationMethod.JupyterHub,
                )

    # ---- Public API
    # -------------------------------------------------------------------------
    def reset_page(self, clear=False):
        """Reset page to allow users to introduce a new connection."""
        # Set a new host id
        self.host_id = str(uuid.uuid4())

        if clear:
            # Reset tracked widgets
            self.reset_widget_dicts()

            # Add a new, clean set of widgets to the page
            ssh_clean_info_widget = self.create_ssh_connection_info_widget()
            self.ssh_widget.layout().addWidget(ssh_clean_info_widget)
            self.ssh_widget.layout().setCurrentWidget(ssh_clean_info_widget)

            jupyterhub_clean_info_widget = (
                self.create_jupyterhub_connection_info_widget()
            )
            self.jupyterhub_widget.layout().addWidget(
                jupyterhub_clean_info_widget
            )
            self.jupyterhub_widget.layout().setCurrentWidget(
                jupyterhub_clean_info_widget
            )
        else:
            # Change option names associated to all widgets present in the page
            # to reference the new host_id
            for widgets in [self.comboboxes, self.lineedits, self.spinboxes]:
                for widget in widgets:
                    section, option, default = widgets[widget]
                    new_option = "/".join(
                        [self.host_id] + option.split("/")[1:]
                    )
                    widgets[widget] = (section, new_option, default)

    def get_current_tab(self, index: int | None = None) -> str:
        if index is None:
            index = self.tabs.currentIndex()

        if index == 0:
            return "SSH"
        else:
            return "JupyterHub"

    def show_ssh_info_widget(self):
        self.ssh_widget.setCurrentWidget(self.ssh_info_widget)

    def show_env_creation_widget(self):
        self.ssh_widget.setCurrentWidget(self.env_creation_widget)

    def show_env_packages_widget(self):
        self.ssh_widget.setCurrentWidget(self.env_packages_widget)

    def is_ssh_info_widget_shown(self) -> bool:
        return self.ssh_widget.currentWidget() == self.ssh_info_widget

    def is_env_creation_widget_shown(self) -> bool:
        return self.ssh_widget.currentWidget() == self.env_creation_widget

    def is_env_packages_widget_shown(self) -> bool:
        return self.ssh_widget.currentWidget() == self.env_packages_widget

    def selected_env_creation_method(self) -> CreateEnvMethods:
        return self.env_method_group.checkedId()

    def validate_env_creation(self):
        method_id = self.env_method_group.checkedId()
        if method_id != CreateEnvMethods.NoEnv:
            env_method_widget = self._radio_buttons_to_info_widgets[method_id]
            if env_method_widget.validate_contents(env_names=[]):
                return True
            else:
                return False

        return True

    def get_create_env_info(self):
        method_id = self.env_method_group.checkedId()
        env_method_widget = self._radio_buttons_to_info_widgets[method_id]
        if method_id == CreateEnvMethods.NewEnv:
            return (
                env_method_widget.get_env_name(),
                env_method_widget.get_python_version()
            )
        elif method_id == CreateEnvMethods.ImportEnv:
            return (
                env_method_widget.get_zip_file(),
                env_method_widget.get_env_name()
            )

    def get_env_packages_list(self):
        return self._packages_info.get_changed_packages()

    def setup_env_packages_widget(self):
        env_name, python_version = self.get_create_env_info()
        self._packages_info.setup(
            env_name,
            python_version,
            f"~/.envs-manager/backends/pixi/{env_name}"
        )

    # ---- Private API
    # -------------------------------------------------------------------------
    def _create_env_creation_widget(self):
        # Intro text
        intro_label = QLabel(
            _("Create a Python environment on the remote host")
        )
        intro_tip_text = _(
            "Decide whether you want to create a remote environment to run "
            "your code and how to do it"
        )
        intro_tip = TipWidget(
            tip_text=intro_tip_text,
            icon=ima.icon('info_tip'),
            hover_icon=ima.icon('info_tip_hover'),
            size=AppStyle.ConfigPageIconSize + 2,
            wrap_text=True,
        )

        # Increase font size to make it more relevant
        font = self.get_font(SpyderFontType.Interface)
        font.setPointSize(font.pointSize() + 1)
        intro_label.setFont(font)

        # Layout
        intro_layout = QHBoxLayout()
        intro_layout.setContentsMargins(0, 0, 0, 0)
        intro_layout.setSpacing(0)
        intro_layout.setAlignment(Qt.AlignCenter)
        intro_layout.addWidget(intro_label)
        intro_layout.addWidget(intro_tip)

        # Available methods
        methods_group = QGroupBox(_("Available methods"))

        self.env_method_group.idToggled.connect(
            self._on_env_creation_method_changed
        )

        new_env_radio = self.create_radiobutton(
            _("Create a new environment"),
            option=None,
            button_group=self.env_method_group,
            id_=CreateEnvMethods.NewEnv,
        )
        import_env_radio = self.create_radiobutton(
            _("Import an existing environment"),
            option=None,
            button_group=self.env_method_group,
            id_=CreateEnvMethods.ImportEnv,
        )
        no_env_radio = self.create_radiobutton(
            _("Don't create an environment"),
            option=None,
            button_group=self.env_method_group,
            id_=CreateEnvMethods.NoEnv,
        )

        methods_layout = QVBoxLayout()
        methods_layout.addSpacing(3)
        methods_layout.addWidget(new_env_radio)
        methods_layout.addWidget(import_env_radio)
        methods_layout.addWidget(no_env_radio)
        methods_group.setLayout(methods_layout)

        # Required info
        info_group = QGroupBox(_("Required information"))

        new_env_info = NewEnvironment(
            self,
            max_width_for_content=470,
            show_in_remote_connections_dialog=True
        )
        new_env_info.setMaximumWidth(470)

        import_env_info = NewEnvironment(
            self,
            max_width_for_content=470,
            import_env=True,
            show_in_remote_connections_dialog=True
        )
        import_env_info.setMaximumWidth(470)

        no_env_info = QLabel(
            _(
                "You can set up an environment later by going to the menu "
                "entry <i>Tools > Environment manager</i>."
            )
        )
        no_env_info.setWordWrap(True)

        info_layout = QVBoxLayout()
        info_layout.addWidget(new_env_info)
        info_layout.addWidget(import_env_info)
        info_layout.addWidget(no_env_info)
        info_group.setLayout(info_layout)

        # Hide all info widgets to only show the one that's checked
        for widget in [new_env_info, import_env_info, no_env_info]:
            widget.setVisible(False)

        # Use the following mapping to show/hide info widgets when the
        # corresponding radio button is toggled
        self._radio_buttons_to_info_widgets = {
            CreateEnvMethods.NewEnv: new_env_info,
            CreateEnvMethods.ImportEnv: import_env_info,
            CreateEnvMethods.NoEnv: no_env_info,
        }

        # Set new env as the default method
        new_env_radio.radiobutton.setChecked(True)

        # Final layout
        layout = QVBoxLayout()
        layout.setContentsMargins(
            3 * AppStyle.MarginSize, 0, 3 * AppStyle.MarginSize, 0
        )
        layout.addLayout(intro_layout)
        layout.addSpacing(8 * AppStyle.MarginSize)
        layout.addWidget(methods_group)
        layout.addWidget(info_group)
        layout.addStretch()

        env_creation_widget = QWidget(self)
        env_creation_widget.setLayout(layout)

        return env_creation_widget

    def _on_env_creation_method_changed(
        self, id_: CreateEnvMethods, checked: bool
    ):
        self._radio_buttons_to_info_widgets[id_].setVisible(checked)

    def _create_env_packages_widget(self):
        # Intro text
        intro_label = QLabel(_("Select packages for your remote environment"))
        intro_tip_text = _(
            "Choose the packages you want to install in your remote Python "
            "environment"
        )
        intro_tip = TipWidget(
            tip_text=intro_tip_text,
            icon=ima.icon("info_tip"),
            hover_icon=ima.icon("info_tip_hover"),
            size=AppStyle.ConfigPageIconSize + 2,
            wrap_text=True,
        )

        # Increase font size to make it more relevant
        font = self.get_font(SpyderFontType.Interface)
        font.setPointSize(font.pointSize() + 1)
        intro_label.setFont(font)

        # Layout
        intro_layout = QHBoxLayout()
        intro_layout.setContentsMargins(0, 0, 0, 0)
        intro_layout.setSpacing(0)
        intro_layout.setAlignment(Qt.AlignCenter)
        intro_layout.addWidget(intro_label)
        intro_layout.addWidget(intro_tip)

        self._packages_info = EditEnvironment(
            self, show_in_remote_connections_dialog=True
        )
        self._packages_info.set_empty_message_visible(True)
        self._packages_info.setMaximumWidth(
            525 if MAC else (485 if WIN else 500)
        )

        # Final layout
        layout = QVBoxLayout()
        layout.setContentsMargins(
            3 * AppStyle.MarginSize,
            0,
            3 * AppStyle.MarginSize,
            # Add bottom margin to let the packages table take the available
            # vertical space
            (2 if MAC else (3 if WIN else 4)) * AppStyle.MarginSize,
        )
        layout.addLayout(intro_layout)
        layout.addSpacing(8 * AppStyle.MarginSize)
        layout.addWidget(self._packages_info)

        env_packages_widget = QWidget(self)
        env_packages_widget.setLayout(layout)

        return env_packages_widget


class ConnectionPage(BaseConnectionPage):
    """Page to display connection status and info for a remote machine."""

    MAX_WIDTH = 600 if MAC else 580

    def __init__(self, parent, host_id):
        super().__init__(parent, host_id)
        self.new_name = None

    # ---- SidebarPage API
    # -------------------------------------------------------------------------
    def get_name(self):
        return self.get_option(f"{self.host_id}/{self.auth_method()}/name")

    def setup_page(self):
        if self.get_client_type() == ClientType.SSH:
            info_widget = self.create_ssh_connection_info_widget()
        else:
            info_widget = self.create_jupyterhub_connection_info_widget()

        self.status_widget = ConnectionStatusWidget(self, self.host_id)

        self.create_tab(_("Connection status"), self.status_widget)
        self.create_tab(_("Connection info"), info_widget)

    def get_icon(self):
        return self.create_icon("remote_server")

    # ---- Public API
    # -------------------------------------------------------------------------
    def save_server_id(self):
        servers = self.get_option("servers", default={})
        servers[self.host_id] = {}
        self.set_option("servers", servers)

    def remove_config_options(self):
        """Remove config options associated to this connection."""
        # Remove current server from the dict of them
        servers = self.get_option("servers")
        servers.pop(self.host_id)
        self.set_option("servers", servers)

        # Remove regular options
        options = [
            "auth_method",
            "password_login/name",
            "password_login/address",
            "password_login/port",
            "password_login/username",
            "keyfile_login/name",
            "keyfile_login/address",
            "keyfile_login/port",
            "keyfile_login/username",
            "configfile_login/name",
            "jupyterhub_login/name",
            "keyfile",
            "configfile",
            "status",
            "status_message",
            "url",
            "client_type",
        ]
        for option in options:
            self.remove_option(f"{self.host_id}/{option}")

        # Remove secure options
        for secure_option in ["password", "passphrase", "token"]:
            # One of these options was saved securely and other as empty in our
            # config system, so we try to remove them both.
            for secure in [True, False]:
                self.remove_option(
                    f"{self.host_id}/{secure_option}", secure=secure
                )

    def update_status(self, info: ConnectionInfo):
        if info["id"] == self.host_id:
            self.status = info["status"]
            self.status_widget.update_status(info)

    def add_log(self, log: RemoteClientLog):
        if log["id"] == self.host_id:
            self.status_widget.add_log(log)

    def add_logs(self, logs: Iterable):
        self.status_widget.add_logs(logs)

    def has_new_name(self):
        """Check if users changed the connection name."""
        current_auth_method = self.auth_method(from_gui=True)
        current_name = self._name_widgets[current_auth_method].textbox.text()

        if self.get_name() != current_name:
            self.new_name = current_name
            return True
        else:
            self.new_name = None
            return False

    def update_connection_info(self):
        self.status_widget.update_info()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016 Pepijn Kenter.
# Copyright (c) 2019- Spyder Project Contributors
#
# Components of objectbrowser originally distributed under
# the MIT (Expat) license. Licensed under the terms of the MIT License;
# see NOTICE.txt in the Spyder root directory for details
# -----------------------------------------------------------------------------

# Standard library imports
import logging
import traceback
from typing import Any, Callable, Optional

# Third-party imports
from qtpy.QtCore import Slot, QModelIndex, QPoint, QSize, Qt
from qtpy.QtGui import QTextOption
from qtpy.QtWidgets import (
    QAbstractItemView, QButtonGroup, QGroupBox, QHBoxLayout, QHeaderView,
    QMessageBox, QPushButton, QRadioButton, QSplitter, QStyle, QToolButton,
    QVBoxLayout, QWidget)

# Local imports
from spyder.api.fonts import SpyderFontsMixin, SpyderFontType
from spyder.api.translations import _
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.plugins.variableexplorer.widgets.basedialog import BaseDialog
from spyder.plugins.variableexplorer.widgets.objectexplorer import (
    DEFAULT_ATTR_COLS, DEFAULT_ATTR_DETAILS, ToggleColumnTreeView,
    TreeItem, TreeModel, TreeProxyModel)
from spyder.utils.icon_manager import ima
from spyder.utils.qthelpers import qapplication
from spyder.utils.stylesheet import AppStyle, MAC
from spyder.widgets.simplecodeeditor import SimpleCodeEditor


logger = logging.getLogger(__name__)


class ObjectExplorerActions:
    Close = "close"
    Refresh = 'refresh_action'
    ShowCallable = 'show_callable_action'
    ShowSpecialAttributes = 'show_special_attributes_action'


class ObjectExplorerMenus:
    Options = 'options_menu'


class ObjectExplorerOptionsMenuSections:
    Toggle = 'toggle_section'
    Close = 'close_section'


class ObjectExplorerWidgets:
    OptionsToolButton = 'options_button_widget'
    Toolbar = 'toolbar'
    ToolbarStretcher = 'toolbar_stretcher'

# About message
EDITOR_NAME = 'Object'


class ObjectExplorer(BaseDialog, SpyderWidgetMixin, SpyderFontsMixin):
    """Object explorer main widget window."""
    CONF_SECTION = 'variable_explorer'

    def __init__(self,
                 obj,
                 name='',
                 expanded=False,
                 resize_to_contents=True,
                 parent=None,
                 namespacebrowser=None,
                 data_function: Optional[Callable[[], Any]] = None,
                 attribute_columns=DEFAULT_ATTR_COLS,
                 attribute_details=DEFAULT_ATTR_DETAILS,
                 readonly=None,
                 reset=False):
        """
        Constructor

        :param obj: any Python object or variable
        :param name: name of the object as it will appear in the root node
        :param expanded: show the first visible root element expanded
        :param resize_to_contents: resize columns to contents ignoring width
            of the attributes
        :param namespacebrowser: the NamespaceBrowser that the object
            originates from, if any
        :param attribute_columns: list of AttributeColumn objects that
            define which columns are present in the table and their defaults
        :param attribute_details: list of AttributeDetails objects that define
            which attributes can be selected in the details pane.
        :param reset: If true the persistent settings, such as column widths,
            are reset.
        """
        super().__init__(parent)
        self.setAttribute(Qt.WA_DeleteOnClose)

        # Model
        self.name = name
        self.expanded = expanded
        self.namespacebrowser = namespacebrowser
        self.data_function = data_function
        self._attr_cols = attribute_columns
        self._attr_details = attribute_details
        self.readonly = readonly

        self.obj_tree = None
        self._proxy_tree_model = None
        self.btn_save_and_close = None
        self.btn_close = None

        # Views
        self._setup_toolbar()
        self._setup_views()
        if self.name:
            self.setWindowTitle(f'{self.name} - {EDITOR_NAME}')
        else:
            self.setWindowTitle(EDITOR_NAME)
        self.setWindowFlags(Qt.Window)

        # Load object into editor
        self.set_value(obj)

        self._resize_to_contents = resize_to_contents
        self._readViewSettings(reset=reset)

    def get_value(self):
        """Get editor current object state."""
        return self._tree_model.inspectedItem.obj

    def set_value(self, obj):
        """Set object displayed in the editor."""
        self._tree_model = TreeModel(obj, obj_name=self.name,
                                     attr_cols=self._attr_cols)

        show_callable_attributes = self.get_conf('show_callable_attributes')
        show_special_attributes = self.get_conf('show_special_attributes')
        self._proxy_tree_model = TreeProxyModel(
            show_callable_attributes=show_callable_attributes,
            show_special_attributes=show_special_attributes
        )

        self._proxy_tree_model.setSourceModel(self._tree_model)
        # self._proxy_tree_model.setSortRole(RegistryTableModel.SORT_ROLE)
        self._proxy_tree_model.setDynamicSortFilter(True)
        # self._proxy_tree_model.setSortCaseSensitivity(Qt.CaseInsensitive)

        # Tree widget
        old_obj_tree = self.obj_tree
        self.obj_tree = ToggleColumnTreeView(
            self.namespacebrowser,
            self.data_function
        )
        self.obj_tree.setAlternatingRowColors(True)
        self.obj_tree.setModel(self._proxy_tree_model)
        self.obj_tree.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.obj_tree.setUniformRowHeights(True)
        self.obj_tree.add_header_context_menu()

        # Keep a temporary reference of the selection_model to prevent
        # segfault in PySide.
        # See http://permalink.gmane.org/gmane.comp.lib.qt.pyside.devel/222
        selection_model = self.obj_tree.selectionModel()
        selection_model.currentChanged.connect(self._update_details)

        # Check if the values of the model have been changed
        self._proxy_tree_model.sig_setting_data.connect(
            self.save_and_close_enable)

        self._proxy_tree_model.sig_update_details.connect(
            self._update_details_for_item)

        # Select first row so that a hidden root node will not be selected.
        first_row_index = self._proxy_tree_model.firstItemIndex()
        self.obj_tree.setCurrentIndex(first_row_index)
        if self._tree_model.inspectedNodeIsVisible or self.expanded:
            self.obj_tree.expand(first_row_index)

        # Stretch last column?
        # It doesn't play nice when columns are hidden and then shown again.
        obj_tree_header = self.obj_tree.header()
        obj_tree_header.setSectionsMovable(True)
        obj_tree_header.setStretchLastSection(False)

        # Add menu item for toggling columns to the Options menu
        for action in self.obj_tree.toggle_column_actions_group.actions():
            self.add_item_to_menu(
                action,
                menu=self.show_cols_submenu,
                section=ObjectExplorerOptionsMenuSections.Toggle,
                before_section=ObjectExplorerOptionsMenuSections.Close
            )
        column_visible = [col.col_visible for col in self._attr_cols]
        for idx, visible in enumerate(column_visible):
            elem = self.obj_tree.toggle_column_actions_group.actions()[idx]
            elem.setChecked(visible)

        # Place tree widget in editor
        if old_obj_tree:
            self.central_splitter.replaceWidget(0, self.obj_tree)
            old_obj_tree.deleteLater()
        else:
            self.central_splitter.insertWidget(0, self.obj_tree)
            self.central_splitter.setCollapsible(0, False)
            self.central_splitter.setCollapsible(1, True)
            self.central_splitter.setSizes([500, 320])

    def _make_show_column_function(self, column_idx):
        """Creates a function that shows or hides a column."""
        show_column = lambda checked: self.obj_tree.setColumnHidden(
            column_idx, not checked)
        return show_column

    def _setup_toolbar(self, show_callable_attributes=False,
                       show_special_attributes=False):
        """
        Sets up the toolbar and the actions in it.
        """
        def do_nothing():
            # .create_action() needs a toggled= parameter, but we can only
            # set it later in the set_value method, so we use this function as
            # a placeholder here.
            pass

        self.toggle_show_callable_action = self.create_action(
            name=ObjectExplorerActions.ShowCallable,
            text=_("Show callable attributes"),
            icon=ima.icon("class"),
            tip=_("Shows/hides attributes that are callable "
                  "(functions, methods etc)"),
            toggled=self._show_callable_attributes,
            option='show_callable_attributes',
            register_action=False
        )

        self.toggle_show_special_attribute_action = self.create_action(
            name=ObjectExplorerActions.ShowSpecialAttributes,
            text=_("Show __special__ attributes"),
            icon=ima.icon("private2"),
            tip=_("Shows or hides __special__ attributes"),
            toggled=self._show_special_attributes,
            option='show_special_attributes',
            register_action=False
        )

        stretcher = self.create_stretcher(
            ObjectExplorerWidgets.ToolbarStretcher
        )

        self.refresh_action = self.create_action(
            name=ObjectExplorerActions.Refresh,
            text=_('Refresh editor with current value of variable in console'),
            icon=ima.icon('refresh'),
            triggered=self.refresh_editor,
            register_action=False
        )

        self.close_action = self.create_action(
            name=ObjectExplorerActions.Close,
            icon=self.create_icon('close_pane'),
            text=_('Close'),
            triggered=self.reject,
            shortcut=self.get_shortcut(ObjectExplorerActions.Close),
            register_action=False,
            register_shortcut=True
        )
        self.register_shortcut_for_widget(name='close', triggered=self.reject)

        self.refresh_action.setEnabled(self.data_function is not None)

        self.show_cols_submenu = self.create_menu(
            ObjectExplorerMenus.Options,
            register=False
        )
        for item in [self.close_action]:
            self.add_item_to_menu(
                item,
                menu=self.show_cols_submenu,
                section=ObjectExplorerOptionsMenuSections.Close
            )

        self.options_button = self.create_toolbutton(
            name=ObjectExplorerWidgets.OptionsToolButton,
            text=_('Options'),
            icon=ima.icon('tooloptions'),
            register=False
        )
        self.options_button.setPopupMode(QToolButton.InstantPopup)
        self.options_button.setMenu(self.show_cols_submenu)

        self.toolbar = self.create_toolbar(
            ObjectExplorerWidgets.Toolbar,
            register=False
        )

        for item in [
            self.toggle_show_callable_action,
            self.toggle_show_special_attribute_action,
            stretcher,
            self.refresh_action,
            self.options_button
        ]:
            self.add_item_to_toolbar(item, self.toolbar)

        self.toolbar.render()

    def _show_callable_attributes(self, value: bool):
        """
        Called when user toggles "show special attributes" option.
        """
        if self._proxy_tree_model:
            self._proxy_tree_model.setShowCallables(value)
        if self.obj_tree:
            self.obj_tree.resize_columns_to_contents()

    def _show_special_attributes(self, value: bool):
        """
        Called when user toggles "show callable attributes" option.
        """
        if self._proxy_tree_model:
            self._proxy_tree_model.setShowSpecialAttributes(value)
        if self.obj_tree:
            self.obj_tree.resize_columns_to_contents()

    def _setup_views(self):
        """Creates the UI widgets."""

        self.central_splitter = QSplitter(self, orientation=Qt.Vertical)

        # Bottom pane
        bottom_pane_widget = QWidget()
        bottom_pane_widget.setContentsMargins(0, 2*AppStyle.MarginSize, 0, 0)
        bottom_layout = QHBoxLayout()
        bottom_layout.setSpacing(0)
        bottom_layout.setContentsMargins(0, 0, 0, 0)
        bottom_pane_widget.setLayout(bottom_layout)
        self.central_splitter.addWidget(bottom_pane_widget)

        group_box = QGroupBox(_("Details"))
        group_box.setStyleSheet(
            'QGroupBox {margin-bottom: 0px; margin-right: -2px;}'
        )
        bottom_layout.addWidget(group_box)

        h_group_layout = QHBoxLayout()
        top_margin = self.style().pixelMetric(QStyle.PM_LayoutTopMargin)
        h_group_layout.setContentsMargins(0, top_margin, 0, 0)
        group_box.setLayout(h_group_layout)

        # Radio buttons
        radio_widget = QWidget()
        radio_layout = QVBoxLayout()
        radio_layout.setContentsMargins(0, 0, 0, 0)  # left top right bottom
        radio_widget.setLayout(radio_layout)

        self.button_group = QButtonGroup(self)
        for button_id, attr_detail in enumerate(self._attr_details):
            radio_button = QRadioButton(attr_detail.name)
            radio_layout.addWidget(radio_button)
            self.button_group.addButton(radio_button, button_id)

        self.button_group.idClicked.connect(
            self._change_details_field)
        self.button_group.button(0).setChecked(True)

        radio_layout.addStretch(1)
        h_group_layout.addWidget(radio_widget)

        # Editor widget
        self.editor = SimpleCodeEditor(self)
        self.editor.setReadOnly(True)
        h_group_layout.addWidget(self.editor)

        # Save and close buttons
        btn_layout = QHBoxLayout()
        btn_layout.addStretch()

        if not self.readonly:
            self.btn_save_and_close = QPushButton(_('Save and Close'))
            self.btn_save_and_close.setDisabled(True)
            self.btn_save_and_close.clicked.connect(self.accept)
            btn_layout.addWidget(self.btn_save_and_close)

        self.btn_close = QPushButton(_('Close'))
        self.btn_close.setAutoDefault(True)
        self.btn_close.setDefault(True)
        self.btn_close.clicked.connect(self.reject)
        btn_layout.addWidget(self.btn_close)

        layout = QVBoxLayout()
        layout.addWidget(self.toolbar)

        # Remove vertical space between toolbar and data from object
        style = self.style()
        default_spacing = style.pixelMetric(QStyle.PM_LayoutVerticalSpacing)
        layout.addSpacing(-default_spacing)

        layout.addWidget(self.central_splitter)
        layout.addSpacing((-1 if MAC else 2) * AppStyle.MarginSize)
        layout.addLayout(btn_layout)
        self.setLayout(layout)

    # End of setup_methods
    def _readViewSettings(self, reset=False):
        """
        Reads the persistent program settings.

        :param reset: If True, the program resets to its default settings.
        """
        pos = QPoint(20, 20)
        window_size = QSize(825, 650)
        details_button_idx = 0

        header = self.obj_tree.header()
        header_restored = False

        if reset:
            logger.debug("Resetting persistent view settings")
        else:
            pos = pos
            window_size = window_size
            details_button_idx = details_button_idx
#            splitter_state = settings.value("central_splitter/state")
            splitter_state = None
            if splitter_state:
                self.central_splitter.restoreState(splitter_state)
#            header_restored = self.obj_tree.read_view_settings(
#                'table/header_state',
#                settings, reset)
            header_restored = False

        if not header_restored:
            column_sizes = [col.width for col in self._attr_cols]
            for idx, size in enumerate(column_sizes):
                if not self._resize_to_contents and size > 0:  # Just in case
                    header.resizeSection(idx, size)
                else:
                    header.resizeSections(QHeaderView.ResizeToContents)
                    break

        self.resize(window_size)

        button = self.button_group.button(details_button_idx)
        if button is not None:
            button.setChecked(True)

    def refresh_editor(self) -> None:
        """
        Refresh data in editor.
        """
        assert self.data_function is not None

        try:
            data = self.data_function()
        except (IndexError, KeyError):
            QMessageBox.critical(
                self,
                _('Object explorer'),
                _('The variable no longer exists.')
            )
            self.reject()
            return

        self.set_value(data)

    @Slot()
    def save_and_close_enable(self):
        """Handle the data change event to enable the save and close button."""
        if self.btn_save_and_close:
            self.btn_save_and_close.setEnabled(True)
            self.btn_save_and_close.setAutoDefault(True)
            self.btn_save_and_close.setDefault(True)

    @Slot(QModelIndex, QModelIndex)
    def _update_details(self, current_index, _previous_index):
        """Shows the object details in the editor given an index."""
        tree_item = self._proxy_tree_model.treeItem(current_index)
        self._update_details_for_item(tree_item)

    def _change_details_field(self, _button_id=None):
        """Changes the field that is displayed in the details pane."""
        # logger.debug("_change_details_field: {}".format(_button_id))
        current_index = self.obj_tree.selectionModel().currentIndex()
        tree_item = self._proxy_tree_model.treeItem(current_index)
        self._update_details_for_item(tree_item)

    @Slot(TreeItem)
    def _update_details_for_item(self, tree_item):
        """Shows the object details in the editor given an tree_item."""
        try:
            # obj = tree_item.obj
            button_id = self.button_group.checkedId()
            assert button_id >= 0, ("No radio button selected. "
                                    "Please report this bug.")
            attr_details = self._attr_details[button_id]
            data = attr_details.data_fn(tree_item)
            self.editor.setPlainText(data)
            self.editor.setWordWrapMode(attr_details.line_wrap)
            self.editor.setup_editor(
                font=self.get_font(SpyderFontType.MonospaceInterface),
                show_blanks=False,
                color_scheme=self.get_conf('selected', section='appearance'),
                scroll_past_end=False,
            )
            self.editor.set_text(data)

            if attr_details.name == 'Source code':
                self.editor.set_language('Python')
            else:
                self.editor.set_language('Rst')

        except Exception as ex:
            self.editor.setStyleSheet("color: red;")
            stack_trace = traceback.format_exc()
            self.editor.setPlainText("{}\n\n{}".format(ex, stack_trace))
            self.editor.setWordWrapMode(
                QTextOption.WrapAtWordBoundaryOrAnywhere)

    @classmethod
    def create_explorer(cls, *args, **kwargs):
        """
        Creates and shows and ObjectExplorer window.

        The *args and **kwargs will be passed to the ObjectExplorer constructor

        A (class attribute) reference to the browser window is kept to prevent
        it from being garbage-collected.
        """
        object_explorer = cls(*args, **kwargs)
        object_explorer.exec_()
        return object_explorer


# =============================================================================
# Tests
# =============================================================================
def test():
    """Run object editor test"""
    import datetime
    import numpy as np
    from spyder.pil_patch import Image

    app = qapplication()

    data = np.random.randint(1, 256, size=(100, 100)).astype('uint8')
    image = Image.fromarray(data)

    class Foobar(object):
        def __init__(self):
            self.text = "toto"

        def get_text(self):
            return self.text
    foobar = Foobar()
    example = {'str': 'kjkj kj k j j kj k jkj',
               'list': [1, 3, 4, 'kjkj', None],
               'set': {1, 2, 1, 3, None, 'A', 'B', 'C', True, False},
               'dict': {'d': 1, 'a': np.random.rand(10, 10), 'b': [1, 2]},
               'float': 1.2233,
               'array': np.random.rand(10, 10),
               'image': image,
               'date': datetime.date(1945, 5, 8),
               'datetime': datetime.datetime(1945, 5, 8),
               'foobar': foobar}
    ObjectExplorer.create_explorer(example, 'Example')


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
# ----------------------------------------------------------------------------

"""
Simple code editor with syntax highlighting and line number area.

Adapted from:
https://doc.qt.io/qt-5/qtwidgets-widgets-codeeditor-example.html
"""

# Third party imports
import qstylizer.style
from qtpy.QtCore import QPoint, QRect, QSize, Qt, Signal
from qtpy.QtGui import (
    QColor,
    QFont,
    QPainter,
    QTextCursor,
    QTextFormat,
    QTextOption,
)
from qtpy.QtWidgets import QPlainTextEdit, QTextEdit, QWidget

# Local imports
import spyder.utils.syntaxhighlighters as sh
from spyder.widgets.mixins import BaseEditMixin


# Constants
LANGUAGE_EXTENSIONS = {
    'Python': ('py', 'pyw', 'python', 'ipy'),
    'Cython': ('pyx', 'pxi', 'pxd'),
    'Enaml': ('enaml',),
    'Fortran77': ('f', 'for', 'f77'),
    'Fortran': ('f90', 'f95', 'f2k', 'f03', 'f08'),
    'Idl': ('pro',),
    'Diff': ('diff', 'patch', 'rej'),
    'GetText': ('po', 'pot'),
    'Nsis': ('nsi', 'nsh'),
    'Html': ('htm', 'html'),
    'Cpp': ('c', 'cc', 'cpp', 'cxx', 'h', 'hh', 'hpp', 'hxx'),
    'OpenCL': ('cl',),
    'Yaml': ('yaml', 'yml'),
    'Markdown': ('md', 'mdw'),
    # Every other language
    'None': ('', ),
}


class LineNumberArea(QWidget):
    """
    Adapted from:
    https://doc.qt.io/qt-5/qtwidgets-widgets-codeeditor-example.html
    """

    def __init__(self, code_editor=None):
        super().__init__(code_editor)

        self._editor = code_editor
        self._left_padding = 6  # Pixels
        self._right_padding = 3  # Pixels

    # --- Qt overrides
    # ------------------------------------------------------------------------
    def sizeHint(self):
        return QSize(self._editor.linenumberarea_width(), 0)

    def paintEvent(self, event):
        self._editor.linenumberarea_paint_event(event)


class SimpleCodeEditor(QPlainTextEdit, BaseEditMixin):
    """Simple editor with highlight features."""

    LANGUAGE_HIGHLIGHTERS = {
        'Python': (sh.PythonSH, '#'),
        'Cython': (sh.CythonSH, '#'),
        'Fortran77': (sh.Fortran77SH, 'c'),
        'Fortran': (sh.FortranSH, '!'),
        'Idl': (sh.IdlSH, ';'),
        'Diff': (sh.DiffSH, ''),
        'GetText': (sh.GetTextSH, '#'),
        'Nsis': (sh.NsisSH, '#'),
        'Html': (sh.HtmlSH, ''),
        'Yaml': (sh.YamlSH, '#'),
        'Cpp': (sh.CppSH, '//'),
        'OpenCL': (sh.OpenCLSH, '//'),
        'Enaml': (sh.EnamlSH, '#'),
        'Markdown': (sh.MarkdownSH, '#'),
        # Every other language
        'None': (sh.TextSH, ''),
    }

    # --- Signals
    # ------------------------------------------------------------------------
    sig_focus_changed = Signal()
    """
    This signal when the focus of the editor changes, either by a
    `focusInEvent` or `focusOutEvent` event.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        # Variables
        self._linenumber_enabled = None
        self._color_scheme = "spyder/dark"
        self._language = None
        self._blanks_enabled = None
        self._scrollpastend_enabled = None
        self._wrap_mode = None
        self._highlight_current_line = None
        self.supported_language = False

        # Widgets
        self._highlighter = None
        self.linenumberarea = LineNumberArea(self)

        # Widget setup
        self.update_linenumberarea_width(0)
        self._apply_current_line_highlight()

        # Style
        self.css = qstylizer.style.StyleSheet()

        # Signals
        self.blockCountChanged.connect(self.update_linenumberarea_width)
        self.updateRequest.connect(self.update_linenumberarea)
        self.cursorPositionChanged.connect(self._apply_current_line_highlight)

    # --- Private API
    # ------------------------------------------------------------------------
    def _apply_color_scheme(self):
        hl = self._highlighter
        if hl is not None:
            hl.setup_formats(self.font())
            if self._color_scheme is not None:
                hl.set_color_scheme(self._color_scheme)

            self._set_palette(background=hl.get_background_color(),
                              foreground=hl.get_foreground_color())

    def _set_palette(self, background, foreground):
        self.css.QPlainTextEdit.setValues(
            background=background.name(),
            color=foreground.name(),
        )
        self.setStyleSheet(self.css.toString())
        self.rehighlight()

    def _apply_current_line_highlight(self):
        if self._highlighter and self._highlight_current_line:
            extra_selections = []
            selection = QTextEdit.ExtraSelection()
            line_color = self._highlighter.get_currentline_color()
            selection.format.setBackground(line_color)
            selection.format.setProperty(QTextFormat.FullWidthSelection, True)
            selection.cursor = self.textCursor()
            selection.cursor.clearSelection()
            extra_selections.append(selection)

            self.setExtraSelections(extra_selections)
        else:
            self.setExtraSelections([])

    # --- Qt Overrides
    # ------------------------------------------------------------------------
    def focusInEvent(self, event):
        self.sig_focus_changed.emit()
        super().focusInEvent(event)

    def focusOutEvent(self, event):
        self.sig_focus_changed.emit()
        super().focusInEvent(event)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        if self._linenumber_enabled:
            cr = self.contentsRect()
            self.linenumberarea.setGeometry(
                QRect(
                    cr.left(),
                    cr.top(),
                    self.linenumberarea_width(),
                    cr.height(),
                )
            )

    # --- Public API
    # ------------------------------------------------------------------------
    def setup_editor(
        self,
        linenumbers=True,
        color_scheme="spyder/dark",
        language="py",
        font=None,
        show_blanks=False,
        wrap=False,
        highlight_current_line=True,
        scroll_past_end=False,
    ):
        """
        Setup editor options.

        Parameters
        ----------
        color_scheme: str, optional
            Default is "spyder/dark".
        language: str, optional
            Default is "py".
        font: QFont or None
            Default is None.
        show_blanks: bool, optional
            Default is False/
        wrap: bool, optional
            Default is False.
        highlight_current_line: bool, optional
            Default is True.
        scroll_past_end: bool, optional
            Default is False
        """
        if font:
            self.set_font(font)

        self.set_highlight_current_line(highlight_current_line)
        self.set_blanks_enabled(show_blanks)
        self.toggle_line_numbers(linenumbers)
        self.set_scrollpastend_enabled(scroll_past_end)
        self.set_language(language)
        self.set_color_scheme(color_scheme)
        self.toggle_wrap_mode(wrap)

    def set_font(self, font):
        """
        Set the editor font.

        Parameters
        ----------
        font: QFont
            Font to use.
        """
        if font:
            self.setFont(font)
            self._apply_color_scheme()

    def set_color_scheme(self, color_scheme):
        """
        Set the editor color scheme.

        Parameters
        ----------
        color_scheme: str
            Color scheme to use.
        """
        self._color_scheme = color_scheme
        self._apply_color_scheme()

    def set_language(self, language):
        """
        Set current syntax highlighting to use `language`.

        Parameters
        ----------
        language: str or None
            Language name or known extensions.
        """
        sh_class = sh.TextSH
        language = str(language).lower()
        self.supported_language = False
        for (key, value) in LANGUAGE_EXTENSIONS.items():
            if language in (key.lower(), ) + value:
                sh_class, __ = self.LANGUAGE_HIGHLIGHTERS[key]
                self._language = key
                self.supported_language = True

        self._highlighter = sh_class(
            self.document(), self.font(), self._color_scheme)
        self._apply_color_scheme()

    def toggle_line_numbers(self, state):
        """
        Set visibility of line number area

        Parameters
        ----------
        state: bool
            Visible state of the line number area.
        """

        self._linenumber_enabled = state
        self.linenumberarea.setVisible(state)
        self.update_linenumberarea_width(())

    def set_scrollpastend_enabled(self, state):
        """
        Set scroll past end state.

        Parameters
        ----------
        state: bool
            Scroll past end state.
        """
        self._scrollpastend_enabled = state
        self.setCenterOnScroll(state)
        self.setDocument(self.document())

    def toggle_wrap_mode(self, state):
        """
        Set line wrap..

        Parameters
        ----------
        state: bool
            Wrap state.
        """
        self.set_wrap_mode('word' if state else None)

    def set_wrap_mode(self, mode=None):
        """
        Set line wrap mode.

        Parameters
        ----------
        mode: str or None, optional
            "word", or "character". Default is None.
        """
        if mode == 'word':
            wrap_mode = QTextOption.WrapAtWordBoundaryOrAnywhere
        elif mode == 'character':
            wrap_mode = QTextOption.WrapAnywhere
        else:
            wrap_mode = QTextOption.NoWrap

        self.setWordWrapMode(wrap_mode)

    def set_highlight_current_line(self, value):
        """
        Set if the current line is highlighted.

        Parameters
        ----------
        value: bool
            The value of the current line highlight option.
        """
        self._highlight_current_line = value
        self._apply_current_line_highlight()

    def set_blanks_enabled(self, state):
        """
        Show blank spaces.

        Parameters
        ----------
        state: bool
            Blank spaces visibility.
        """
        self._blanks_enabled = state
        option = self.document().defaultTextOption()
        option.setFlags(option.flags()
                        | QTextOption.AddSpaceForLineAndParagraphSeparators)

        if self._blanks_enabled:
            option.setFlags(option.flags() | QTextOption.ShowTabsAndSpaces)
        else:
            option.setFlags(option.flags() & ~QTextOption.ShowTabsAndSpaces)

        self.document().setDefaultTextOption(option)

        # Rehighlight to make the spaces less apparent.
        self.rehighlight()

    # --- Line number area
    # ------------------------------------------------------------------------
    def linenumberarea_paint_event(self, event):
        """
        Paint the line number area.
        """
        if self._linenumber_enabled:
            painter = QPainter(self.linenumberarea)
            painter.fillRect(
                event.rect(),
                self._highlighter.get_sideareas_color(),
            )

            block = self.firstVisibleBlock()
            block_number = block.blockNumber()
            top = round(self.blockBoundingGeometry(block).translated(
                self.contentOffset()).top())
            bottom = top + round(self.blockBoundingRect(block).height())

            font = self.font()
            active_block = self.textCursor().block()
            active_line_number = active_block.blockNumber() + 1

            while block.isValid() and top <= event.rect().bottom():
                if block.isVisible() and bottom >= event.rect().top():
                    number = block_number + 1

                    if number == active_line_number:
                        font.setWeight(QFont.Weight.Bold)
                        painter.setFont(font)
                        painter.setPen(
                            self._highlighter.get_foreground_color())
                    else:
                        font.setWeight(QFont.Weight.Normal)
                        painter.setFont(font)
                        painter.setPen(QColor(Qt.darkGray))
                    right_padding = self.linenumberarea._right_padding
                    painter.drawText(
                        0,
                        top,
                        self.linenumberarea.width() - right_padding,
                        self.fontMetrics().height(),
                        Qt.AlignRight, str(number),
                    )

                block = block.next()
                top = bottom
                bottom = top + round(self.blockBoundingRect(block).height())
                block_number += 1

    def linenumberarea_width(self):
        """
        Return the line number area width.

        Returns
        -------
        int
            Line number are width in pixels.

        Notes
        -----
        If the line number area is disabled this will return zero.
        """
        width = 0
        if self._linenumber_enabled:
            digits = 1
            count = max(1, self.blockCount())
            while count >= 10:
                count /= 10
                digits += 1

            fm = self.fontMetrics()
            width = (self.linenumberarea._left_padding
                     + self.linenumberarea._right_padding
                     + fm.width('9') * digits)

        return width

    def update_linenumberarea_width(self, new_block_count=None):
        """
        Update the line number area width based on the number of blocks in
        the document.

        Parameters
        ----------
        new_block_count: int
            The current number of blocks in the document.
        """
        self.setViewportMargins(self.linenumberarea_width(), 0, 0, 0)

    def update_linenumberarea(self, rect, dy):
        """
        Update scroll position of line number area.
        """
        if self._linenumber_enabled:
            if dy:
                self.linenumberarea.scroll(0, dy)
            else:
                self.linenumberarea.update(
                    0, rect.y(), self.linenumberarea.width(), rect.height())

            if rect.contains(self.viewport().rect()):
                self.update_linenumberarea_width(0)

    # --- Text and cursor handling
    # ------------------------------------------------------------------------
    def set_selection(self, start, end):
        """
        Set current text selection.

        Parameters
        ----------
        start: int
            Selection start position.
        end: int
            Selection end position.
        """
        cursor = self.textCursor()
        cursor.setPosition(start)
        cursor.setPosition(end, QTextCursor.KeepAnchor)
        self.setTextCursor(cursor)

    def stdkey_backspace(self):
        if not self.has_selected_text():
            self.moveCursor(QTextCursor.PreviousCharacter,
                            QTextCursor.KeepAnchor)
        self.remove_selected_text()

    def restrict_cursor_position(self, position_from, position_to):
        """
        Restrict the cursor from being inside from and to positions.

        Parameters
        ----------
        position_from: int
            Selection start position.
        position_to: int
            Selection end position.
        """
        position_from = self.get_position(position_from)
        position_to = self.get_position(position_to)
        cursor = self.textCursor()
        cursor_position = cursor.position()
        if cursor_position < position_from or cursor_position > position_to:
            self.set_cursor_position(position_to)

    def truncate_selection(self, position_from):
        """
        Restrict the cursor selection to start from the given position.

        Parameters
        ----------
        position_from: int
            Selection start position.
        """
        position_from = self.get_position(position_from)
        cursor = self.textCursor()
        start, end = cursor.selectionStart(), cursor.selectionEnd()
        if start < end:
            start = max([position_from, start])
        else:
            end = max([position_from, end])

        self.set_selection(start, end)

    def set_text(self, text):
        """
        Set `text` of the document.

        Parameters
        ----------
        text: str
            Text to set.
        """
        self.setPlainText(text)

    def append(self, text):
        """
        Add `text` to the end of the document.

        Parameters
        ----------
        text: str
            Text to append.
        """
        cursor = self.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.insertText(text)

    def get_visible_block_numbers(self):
        """Get the first and last visible block numbers."""
        first = self.firstVisibleBlock().blockNumber()
        bottom_right = QPoint(self.viewport().width() - 1,
                              self.viewport().height() - 1)
        last = self.cursorForPosition(bottom_right).blockNumber()
        return (first, last)

    # --- Syntax highlighter
    # ------------------------------------------------------------------------
    def rehighlight(self):
        """
        Reapply syntax highligthing to the document.
        """
        if self._highlighter:
            self._highlighter.rehighlight()
            self._apply_current_line_highlight()


if __name__ == "__main__":
    from spyder.utils.qthelpers import qapplication

    app = qapplication()
    editor = SimpleCodeEditor()
    editor.setup_editor(language="markdown")
    editor.set_text("# Hello!")
    editor.show()
    app.exec_()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2010 IPython Development Team
# Copyright (c) 2013- Spyder Project Contributors
#
# Distributed under the terms of the Modified BSD License
# (BSD 3-clause; see NOTICE.txt in the Spyder root directory for details).
# -----------------------------------------------------------------------------

"""
Calltip widget used only to show signatures.

Adapted from IPython/frontend/qt/console/call_tip_widget.py of the
`IPython Project <https://github.com/ipython/ipython>`_.
Now located at qtconsole/call_tip_widget.py as part of the
`Jupyter QtConsole Project <https://github.com/jupyter/qtconsole>`_.
"""

# Standard library imports
from unicodedata import category
import sys

# Third party imports
import qstylizer.style
from qtpy.QtCore import (QBasicTimer, QCoreApplication, QEvent, Qt, QTimer,
                         Signal)
from qtpy.QtGui import QCursor, QFontMetrics, QPalette
from qtpy.QtWidgets import (QApplication, QFrame, QLabel, QTextEdit,
                            QPlainTextEdit, QStyle, QStyleOptionFrame,
                            QStylePainter, QToolTip)

# Local imports
from spyder.config.gui import is_dark_interface
from spyder.utils.palette import SpyderPalette


BACKGROUND_COLOR = (
    SpyderPalette.COLOR_BACKGROUND_4 if is_dark_interface() else
    SpyderPalette.COLOR_BACKGROUND_2
)


class ToolTipWidget(QLabel):
    """
    Shows tooltips that can be styled with the different themes.
    """

    # Delay in miliseconds before hiding the tooltip
    HIDE_DELAY = 50

    # Signals
    sig_completion_help_requested = Signal(str, str)
    sig_help_requested = Signal(str)

    def __init__(self, parent=None):
        """
        Shows tooltips that can be styled with the different themes.
        """
        super().__init__(parent, Qt.ToolTip)

        # Variables
        self.completion_doc = None
        self._url = ''
        self.app = QCoreApplication.instance()
        self._as_hover = False
        self._as_hint = False
        self._hovered = False
        self._timer_hide = QTimer()
        self._text_edit = parent
        self.show_help_on_click = False

        # Setup
        # This keeps the hints below other applications
        if sys.platform == 'darwin':
            self.setWindowFlags(Qt.SplashScreen)
        else:
            self.setWindowFlags(Qt.ToolTip | Qt.FramelessWindowHint)

        self._timer_hide.setInterval(self.HIDE_DELAY)
        self.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.setTextInteractionFlags(Qt.TextBrowserInteraction)
        self.setOpenExternalLinks(False)
        self.setForegroundRole(QPalette.ToolTipText)
        self.setBackgroundRole(QPalette.ToolTipBase)
        self.setPalette(QToolTip.palette())
        self.setAlignment(Qt.AlignLeft)
        self.setIndent(1)
        self.setFrameStyle(QFrame.NoFrame)
        style = self.style()
        delta_margin = style.pixelMetric(QStyle.PM_ToolTipLabelFrameWidth,
                                         None, self)
        self.setMargin(1 + delta_margin)

        # Signals
        self.linkHovered.connect(self._update_hover_html_link_style)
        self._timer_hide.timeout.connect(self._hide)
        QApplication.instance().applicationStateChanged.connect(
            self._should_hide
        )

        # Style
        self.setStyleSheet(self._stylesheet)

    # ---- Private API
    # -------------------------------------------------------------------------
    def _update_hover_html_link_style(self, url):
        """Update style of labels that include rich text and html links."""
        link = 'text-decoration:none;'
        link_hovered = 'text-decoration:underline;'
        self._url = url

        if url:
            self.setCursor(Qt.PointingHandCursor)
            new_text, old_text = link_hovered, link
        else:
            new_text, old_text = link, link_hovered

        text = self.text()
        new_text = text.replace(old_text, new_text)

        self.setText(new_text)

    def _should_hide(self, state):
        """
        This widget should hide itself if the application is not active.
        """
        if state != Qt.ApplicationActive:
            self._hide()

    def _hide(self):
        """Call the actual hide method."""
        super().hide()

    @property
    def _stylesheet(self):
        css = qstylizer.style.StyleSheet()

        css["ToolTipWidget"].setValues(
            backgroundColor=BACKGROUND_COLOR,
            border=f"1px solid {SpyderPalette.COLOR_TEXT_4}"
        )

        return css.toString()

    # ---- Public API
    # -------------------------------------------------------------------------
    def show_tip(self, point, tip, cursor=None, completion_doc=None,
                 vertical_position='bottom', show_help_on_click=False):
        """Attempt to show the tip at the current mouse location."""

        # Don't show the widget if the main window is not focused
        if QApplication.instance().applicationState() != Qt.ApplicationActive:
            return

        # Set the text and resize the widget accordingly.
        self.setText(tip)
        self.resize(self.sizeHint())

        self.completion_doc = completion_doc
        self.show_help_on_click = show_help_on_click

        text_edit = self._text_edit
        screen_rect = self.screen().geometry()

        tip_height = self.size().height()
        tip_width = self.size().width()

        text_edit_height = text_edit.size().height()
        text_edit_gpos = (
            # If text_edit is a window in itself (e.g. when used indenpendently
            # in our tests) we don't need to get its global position because
            # text_edit.pos() already gives it.
            text_edit.mapToGlobal(text_edit.pos())
            if not text_edit.isWindow()
            else text_edit.pos()
        )

        vertical = vertical_position
        horizontal = 'Right'

        # Check if tip is vertically off text_edit
        if (
            # Tip is off at text_edit's bottom
            point.y() + tip_height > text_edit_gpos.y() + text_edit_height
            # Tip is off at the top
            or point.y() - tip_height < text_edit_gpos.y()
        ):
            if point.y() - tip_height < text_edit_gpos.y():
                vertical = 'bottom'
            else:
                vertical = 'top'

        # Check if tip is horizontally off screen to the right
        if point.x() + tip_width > screen_rect.x() + screen_rect.width():
            if 2 * point.x() < screen_rect.width():
                horizontal = 'Right'
            else:
                horizontal = 'Left'

        # Set coordinates where the tip will be placed
        if vertical == 'top':
            # The +2 below is necessary to leave no vertical space between the
            # tooltip and the text.
            point.setY(point.y() - tip_height + 2)
        else:
            font = text_edit.font()
            text_height = QFontMetrics(font).capHeight()

            # Ubuntu Mono has a strange behavior regarding its height that we
            # need to account for. Other monospaced fonts behave as expected.
            if font.family() == 'Ubuntu Mono':
                padding = 2
            else:
                padding = 1 if sys.platform == "darwin" else 5

            # Qt sets the mouse coordinates (given by `point`) a bit above the
            # line where it's placed, i.e. not exactly where the text
            # vertically starts. So, we need to add an extra height (twice the
            # cap height plus some padding) to place the tip at the line's
            # bottom.
            point.setY(point.y() + 2 * text_height + padding)

        if horizontal == 'Left':
            point.setX(point.x() - tip_width)
        else:
            # The -2 below is necessary to horizontally align the hover to
            # the text.
            point.setX(point.x() - (2 if self._as_hover else 0))

        # Move tip to new coordinates
        self.move(point)

        # Show tip
        if not self.isVisible():
            self._timer_hide.stop()
            self.show()

        return True

    def set_as_tooltip(self):
        """Make the widget work as a tooltip."""
        self.reset_state()

    def set_as_hint(self):
        """Make the widget work to display code completion hints."""
        self._as_hint = True
        self._as_hover = False

    def set_as_hover(self):
        """Make the widget work to display hovers."""
        self._as_hover = True
        self._as_hint = False

    def is_hint(self):
        """Check if the widget is used as a completion hint."""
        return self._as_hint

    def is_hover(self):
        """Check if the widget is used as a hover."""
        return self._as_hover

    def is_hovered(self):
        """Check if the the mouse is on top of this widget."""
        return self._hovered

    def reset_state(self):
        """Reset widget state as a hover, tooltip or hint."""
        self._as_hint = False
        self._as_hover = False

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def paintEvent(self, event):
        """Reimplemented to paint the background."""
        painter = QStylePainter(self)
        option = QStyleOptionFrame()
        option.initFrom(self)
        painter.drawPrimitive(QStyle.PE_PanelTipLabel, option)
        painter.end()

        super().paintEvent(event)

    def mousePressEvent(self, event):
        """Reimplemented to handle mouse press events."""
        if self.completion_doc:
            name = self.completion_doc.get('name', '')
            signature = self.completion_doc.get('signature', '')
            self.sig_completion_help_requested.emit(name, signature)
        elif self.show_help_on_click:
            self.sig_help_requested.emit('')
        elif self._url:
            self.sig_help_requested.emit(self._url)

        super().mousePressEvent(event)

        # Prevent to hide the widget when it's used as a completion hint and
        # users click on it
        if not self._as_hint:
            self._hide()

    def focusOutEvent(self, event):
        """Reimplemented to hide tooltip when focus goes out."""
        self._hide()

    def enterEvent(self, event):
        """Reimplemented to keep tooltip visible and change cursor shape."""
        if self.show_help_on_click:
            self.setCursor(Qt.PointingHandCursor)
        else:
            self.setCursor(Qt.ArrowCursor)

        self._hovered = True
        self._timer_hide.stop()
        super().enterEvent(event)

    def leaveEvent(self, event):
        """Reimplemented to hide tooltip on leave."""
        # This is necessary to set the cursor correctly in enterEvent on Mac
        self.unsetCursor()

        self._hovered = False
        super().leaveEvent(event)

        # Prevent to hide the widget when it's used as a completion hint and
        # the cursor lays on top of it when browsing different completions.
        if not self._as_hint:
            self._hide()

    def hide(self):
        """
        Reimplemented to wait for a little bit before hiding the tooltip.

        This is necessary to leave time to users to hover over the tooltip if
        they want to click on it. If not, the tooltip hides too quickly because
        Qt can emit a hideEvent in the meantime.
        """
        self._timer_hide.start()


class CallTipWidget(QLabel):
    """ Shows call tips by parsing the current text of Q[Plain]TextEdit.
    """

    def __init__(self, text_edit, hide_timer_on=False, as_tooltip=False):
        """ Create a call tip manager that is attached to the specified Qt
            text edit widget.
        """
        assert isinstance(text_edit, (QTextEdit, QPlainTextEdit))
        super().__init__(text_edit, Qt.ToolTip)
        self.app = QCoreApplication.instance()
        self.as_tooltip = as_tooltip

        self.hide_timer_on = hide_timer_on
        self.tip = None
        self._hide_timer = QBasicTimer()
        self._text_edit = text_edit
        self._start_position = -1

        # Setup
        if sys.platform == 'darwin':
            # This keeps the hints below other applications
            self.setWindowFlags(Qt.SplashScreen)
        else:
            self.setWindowFlags(Qt.ToolTip | Qt.FramelessWindowHint)

        self.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.setFont(text_edit.document().defaultFont())
        self.setForegroundRole(QPalette.ToolTipText)
        self.setBackgroundRole(QPalette.ToolTipBase)
        self.setPalette(QToolTip.palette())

        self.setAlignment(Qt.AlignLeft)
        self.setIndent(1)
        self.setFrameStyle(QFrame.NoFrame)
        self.setMargin(1 + self.style().pixelMetric(
                QStyle.PM_ToolTipLabelFrameWidth, None, self))

        # Signals
        QApplication.instance().applicationStateChanged.connect(
            self._should_hide
        )

        # Style
        self.setStyleSheet(self._stylesheet)

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def eventFilter(self, obj, event):
        """ Reimplemented to hide on certain key presses and on text edit focus
            changes.
        """
        if obj == self._text_edit:
            etype = event.type()

            if etype == QEvent.KeyPress:
                key = event.key()
                cursor = self._text_edit.textCursor()
                prev_char = self._text_edit.get_character(cursor.position(),
                                                          offset=-1)
                if key in (Qt.Key_Enter, Qt.Key_Return,
                           Qt.Key_Down, Qt.Key_Up):
                    self.hide()
                elif key == Qt.Key_Escape:
                    self.hide()
                    return True
                elif prev_char == ')':
                    self.hide()

            elif etype == QEvent.FocusOut:
                self.hide()

            elif etype == QEvent.Enter:
                if (self._hide_timer.isActive() and
                  self.app.topLevelAt(QCursor.pos()) == self):
                    self._hide_timer.stop()

            elif etype == QEvent.Leave:
                self._leave_event_hide()

            elif etype == QEvent.WindowBlocked:
                self.hide()

        return super().eventFilter(obj, event)

    def timerEvent(self, event):
        """ Reimplemented to hide the widget when the hide timer fires.
        """
        if event.timerId() == self._hide_timer.timerId():
            self._hide_timer.stop()
            self.hide()

    def enterEvent(self, event):
        """ Reimplemented to cancel the hide timer.
        """
        super().enterEvent(event)
        if self.as_tooltip:
            self.hide()

        if (self._hide_timer.isActive() and
          self.app.topLevelAt(QCursor.pos()) == self):
            self._hide_timer.stop()

    def hideEvent(self, event):
        """ Reimplemented to disconnect signal handlers and event filter.
        """
        super().hideEvent(event)
        # This is needed for issue spyder-ide/spyder#9221,
        try:
            self._text_edit.cursorPositionChanged.disconnect(
                self._cursor_position_changed)
        except (TypeError, RuntimeError):
            pass
        else:
            self._text_edit.removeEventFilter(self)

    def leaveEvent(self, event):
        """ Reimplemented to start the hide timer.
        """
        super().leaveEvent(event)
        self._leave_event_hide()

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        self.hide()

    def paintEvent(self, event):
        """ Reimplemented to paint the background panel.
        """
        painter = QStylePainter(self)
        option = QStyleOptionFrame()
        option.initFrom(self)
        painter.drawPrimitive(QStyle.PE_PanelTipLabel, option)
        painter.end()

        super().paintEvent(event)

    def setFont(self, font):
        """ Reimplemented to allow use of this method as a slot.
        """
        super().setFont(font)

    def showEvent(self, event):
        """ Reimplemented to connect signal handlers and event filter.
        """
        super().showEvent(event)
        self._text_edit.cursorPositionChanged.connect(
            self._cursor_position_changed)
        self._text_edit.installEventFilter(self)

    def focusOutEvent(self, event):
        """ Reimplemented to hide it when focus goes out of the main
            window.
        """
        self.hide()

    # ---- Public API
    # -------------------------------------------------------------------------
    def show_tip(self, point, tip, wrapped_tiplines):
        """ Attempts to show the specified tip at the current cursor location.
        """
        # Don't show the widget if the main window is not focused
        if QApplication.instance().applicationState() != Qt.ApplicationActive:
            return

        # Don't attempt to show it if it's already visible and the text
        # to be displayed is the same as the one displayed before.
        if self.isVisible():
            if self.tip == tip:
                return True
            else:
                self.hide()

        # Attempt to find the cursor position at which to show the call tip.
        text_edit = self._text_edit
        cursor = text_edit.textCursor()
        search_pos = cursor.position() - 1
        self._start_position, _ = self._find_parenthesis(search_pos,
                                                         forward=False)
        if self._start_position == -1:
            return False

        if self.hide_timer_on:
            self._hide_timer.stop()
            # Logic to decide how much time to show the calltip depending
            # on the amount of text present
            if len(wrapped_tiplines) == 1:
                args = wrapped_tiplines[0].split('(')[1]
                nargs = len(args.split(','))
                if nargs == 1:
                    hide_time = 1400
                elif nargs == 2:
                    hide_time = 1600
                else:
                    hide_time = 1800
            elif len(wrapped_tiplines) == 2:
                args1 = wrapped_tiplines[1].strip()
                nargs1 = len(args1.split(','))
                if nargs1 == 1:
                    hide_time = 2500
                else:
                    hide_time = 2800
            else:
                hide_time = 3500
            self._hide_timer.start(hide_time, self)

        # Set the text and resize the widget accordingly.
        self.tip = tip
        self.setText(tip)
        self.resize(self.sizeHint())

        # Locate and show the widget. Place the tip below the current line
        # unless it would be off the window or screen. In that case, decide the
        # best location based trying to minimize the area that goes off-screen.
        cursor_rect = text_edit.cursorRect(cursor)
        screen_rect = self.screen().geometry()
        if self.app.activeWindow():
            window_rect = self.app.activeWindow().geometry()
        else:
            window_rect = screen_rect
        tip_height = self.size().height()
        tip_width = self.size().width()

        vertical = 'bottom'
        horizontal = 'Right'

        if point.y() + tip_height > window_rect.height() + window_rect.y():
            point_ = text_edit.mapToGlobal(cursor_rect.topRight())
            # If tip is still off screen, check if point is in top or bottom
            # half of screen.
            if point_.y() < tip_height:
                # If point is in upper half of screen, show tip below it.
                # otherwise above it.
                if 2 * point.y() < window_rect.height():
                    vertical = 'bottom'
                else:
                    vertical = 'top'
            else:
                vertical = 'top'

        if point.x() + tip_width > screen_rect.width() + screen_rect.x():
            point_ = text_edit.mapToGlobal(cursor_rect.topRight())
            # If tip is still off-screen, check if point is in the right or
            # left half of the screen.
            if point_.x() < tip_width:
                if 2 * point.x() < screen_rect.width():
                    horizontal = 'Right'
                else:
                    horizontal = 'Left'
            else:
                horizontal = 'Left'

        if vertical == 'top':
            point.setY(point.y() - tip_height)
        else:
            font = text_edit.font()
            text_height = QFontMetrics(font).capHeight()

            # Ubuntu Mono has a strange behavior regarding its height that we
            # need to account for. Other monospaced fonts behave as expected.
            if font.family() == 'Ubuntu Mono':
                padding = 0
            else:
                padding = 3

            # Qt sets the mouse coordinates (given by `point`) a bit above the
            # line where it's placed, i.e. not exactly where the text
            # vertically starts. So, we need to add an extra height (twice the
            # cap height plus some padding) to place the tip at the line's
            # bottom.
            point.setY(point.y() + 2 * text_height + padding)

        if horizontal == 'Left':
            point.setX(point.x() - tip_width)

        self.move(point)
        self.show()
        return True

    # ---- Private API
    # -------------------------------------------------------------------------
    def _find_parenthesis(self, position, forward=True):
        """ If 'forward' is True (resp. False), proceed forwards
            (resp. backwards) through the line that contains 'position' until an
            unmatched closing (resp. opening) parenthesis is found. Returns a
            tuple containing the position of this parenthesis (or -1 if it is
            not found) and the number commas (at depth 0) found along the way.
        """
        commas = depth = 0
        document = self._text_edit.document()
        char = str(document.characterAt(position))
        # Search until a match is found or a non-printable character is
        # encountered.
        while category(char) != 'Cc' and position > 0:
            if char == ',' and depth == 0:
                commas += 1
            elif char == ')':
                if forward and depth == 0:
                    break
                depth += 1
            elif char == '(':
                if not forward and depth == 0:
                    break
                depth -= 1
            position += 1 if forward else -1
            char = str(document.characterAt(position))
        else:
            position = -1
        return position, commas

    def _leave_event_hide(self):
        """ Hides the tooltip after some time has passed (assuming the cursor is
            not over the tooltip).
        """
        if (self.hide_timer_on and not self._hide_timer.isActive() and
            # If Enter events always came after Leave events, we wouldn't need
            # this check. But on Mac OS, it sometimes happens the other way
            # around when the tooltip is created.
            self.app.topLevelAt(QCursor.pos()) != self):
            self._hide_timer.start(800, self)

    def _cursor_position_changed(self):
        """ Updates the tip based on user cursor movement.
        """
        cursor = self._text_edit.textCursor()
        position = cursor.position()
        document = self._text_edit.document()
        char = str(document.characterAt(position - 1))
        if position <= self._start_position:
            self.hide()
        elif char == ')':
            pos, _ = self._find_parenthesis(position - 1, forward=False)
            if pos == -1:
                self.hide()

    def _should_hide(self, state):
        """
        This widget should hide itself if the application is not active.
        """
        if state != Qt.ApplicationActive:
            self.hide()

    @property
    def _stylesheet(self):
        css = qstylizer.style.StyleSheet()

        css["CallTipWidget"].setValues(
            backgroundColor=BACKGROUND_COLOR,
            border=f"1px solid {SpyderPalette.COLOR_TEXT_4}"
        )

        return css.toString()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# Third party imports
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QAction
from qtpy import PYSIDE2, PYSIDE6

# Local imports
from spyder.api.plugin_registration.registry import PreferencesAdapter
from spyder.api.translations import _
from spyder.api.widgets.main_container import PluginMainContainer
from spyder.plugins.preferences import MOST_IMPORTANT_PAGES
from spyder.plugins.preferences.api import PreferencesActions
from spyder.plugins.preferences.widgets.configdialog import ConfigDialog


class PreferencesContainer(PluginMainContainer):
    sig_reset_preferences_requested = Signal()
    """Request a reset of preferences."""

    sig_show_preferences_requested = Signal()
    """Request showing preferences."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dialog = None
        self.dialog_index = 0
        self._dialog_size = None

    def create_dialog(self, config_pages, config_tabs, main_window):

        def _dialog_finished(result_code):
            """Restore preferences dialog instance variable."""
            if PYSIDE2 or PYSIDE6:
                self.dialog.disconnect(None, None, None)
            else:
                self.dialog.disconnect()

            self.dialog = None

        if self.dialog is None:
            # TODO: Remove all references to main window
            dlg = ConfigDialog(main_window)
            self.dialog = dlg

            if self._dialog_size is None:
                self._dialog_size = self.get_conf('dialog_size')
            dlg.resize(*self._dialog_size)

            for page_name in config_pages:
                # Add separator before the Plugins page
                if page_name == PreferencesAdapter.NAME:
                    dlg.add_separator()

                (api, ConfigPage, plugin) = config_pages[page_name]
                if api == 'new':
                    page = ConfigPage(plugin, dlg)
                    page.initialize()
                    for Tab in config_tabs.get(page_name, []):
                        page._add_tab(Tab)
                    dlg.add_page(page)
                else:
                    page = plugin._create_configwidget(dlg, main_window)
                    for Tab in config_tabs.get(page_name, []):
                        page._add_tab(Tab)
                    dlg.add_page(page)

                # Add separator after the last element of the most important
                # pages
                if page_name == MOST_IMPORTANT_PAGES[-1]:
                    dlg.add_separator()

            dlg.set_current_index(self.dialog_index)
            dlg.show()
            dlg.check_all_settings()

            dlg.finished.connect(_dialog_finished)
            dlg.pages_widget.currentChanged.connect(
                self.__preference_page_changed)
            dlg.sig_size_changed.connect(self._set_dialog_size)
            dlg.sig_reset_preferences_requested.connect(
                self.sig_reset_preferences_requested)
        else:
            self.dialog.resize(*self._dialog_size)
            self.dialog.show()
            self.dialog.activateWindow()
            self.dialog.raise_()
            self.dialog.setFocus()

    # ---- Private API
    def __preference_page_changed(self, index):
        """Preference page index has changed."""
        self.dialog_index = index

    def _set_dialog_size(self, size):
        self._dialog_size = (size.width(), size.height())

    # ---- Public API
    def is_preferences_open(self):
        """Check if preferences is open."""
        return self.dialog is not None and self.dialog.isVisible()

    def close_preferences(self):
        """Close preferences"""
        if self.dialog is not None:
            self.dialog.close()

    def show_preferences(self):
        """Show preferences."""
        self.sig_show_preferences_requested.emit()

    # ---- PluginMainContainer API
    def setup(self):
        self.show_action = self.create_action(
            PreferencesActions.Show,
            _("Preferences"),
            icon=self.create_icon('configure'),
            triggered=self.show_preferences,
            menurole=QAction.PreferencesRole
        )

        self.reset_action = self.create_action(
            PreferencesActions.Reset,
            _("Reset all preferences to defaults"),
            triggered=self.sig_reset_preferences_requested,
            icon=self.create_icon('reset_factory_defaults'),
        )

    def update_actions(self):
        pass

    def on_close(self):
        # Save dialog size to use it in the next Spyder session
        if isinstance(self._dialog_size, tuple):
            self.set_conf('dialog_size', self._dialog_size)



# -----------------------------------------------------------------------------
# Copyright (c) 2021- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Helper functions to work with the Spyder plugin API.
"""

from __future__ import annotations

from abc import ABCMeta as BaseABCMeta
import collections.abc
import sys
import typing

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec  # noqa: ICN003

_P = ParamSpec("_P")
_T = typing.TypeVar("_T")


def get_class_values(cls: type[object]) -> list[str]:
    """
    Get the attribute values for the class enumerations used in our API.

    Parameters
    ----------
    cls
        Class object to list the enumeration values of.

    Returns
    -------
    list[str]
        String attribute values from a Spyder pseudo-"enum" class.
    """
    return [
        getattr(cls, attr) for attr in dir(cls) if not attr.startswith("_")
    ]


class PrefixNode:
    """Utility class used to represent a prefixed string tuple."""

    def __init__(self, path: tuple[str, ...] | None = None) -> None:
        """
        Representation of a prefixed string tuple.

        Parameters
        ----------
        path : tuple[str, ...] | None, optional
            Underlying prefixed string tuple. The default is None.

        Returns
        -------
        None
        """
        self.children = {}
        self.path = path

    def __iter__(self):
        prefix = [((self.path,), self)]
        while prefix != []:
            current_prefix, node = prefix.pop(0)
            prefix += [
                (current_prefix + (c,), node.children[c])
                for c in node.children
            ]
            yield current_prefix

    def add_path(self, path: tuple[str, ...]) -> None:
        """
        Add a path to the prefix node.

        Parameters
        ----------
        path : tuple[str, ...]
            Underlying prefixed string tuple.

        Returns
        -------
        None
        """
        prefix, *rest = path
        if prefix not in self.children:
            self.children[prefix] = PrefixNode(prefix)

        if len(rest) > 0:
            child = self.children[prefix]
            child.add_path(rest)


class PrefixedTuple(PrefixNode):
    """Utility class to store and iterate over prefixed string tuples."""

    def __iter__(self):
        for key in self.children:
            child = self.children[key]
            for prefix in child:
                yield prefix


class classproperty(property):
    """
    Decorator to declare class constants requiring computation as properties.

    Idea from `Stack Overflow <https://stackoverflow.com/a/7864317/438386>`__.
    """

    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()


class DummyAttribute:
    """Dummy class to mark abstract attributes."""

    pass


def abstract_attribute(
    obj: collections.abc.Callable[_P, _T] | DummyAttribute | None = None,
) -> _T:
    """
    Decorator to mark abstract attributes.

    Must be used in conjunction with the :class:`abc.ABCMeta` metaclass.

    Parameters
    ----------
    obj: collections.abc.Callable[_P, _T] | DummyAttribute | None, optional
        The callable attribute to mark as abstract, a new instance of
        :class:`DummyAttribute` by default.

    Returns
    -------
    _T
        The result of executing the callable attribute ``obj``.
    """
    if obj is None:
        obj = DummyAttribute()
    setattr(obj, "__is_abstract_attribute__", True)
    return obj  # type: ignore


class ABCMeta(BaseABCMeta):
    """
    Metaclass to mark abstract classes.

    Adds support for abstract attributes. If a class has abstract attributes
    and is instantiated, a NotImplementedError is raised.

    Usage
    -----

    .. code-block:: python

        class MyABC(metaclass=ABCMeta):
            @abstract_attribute
            def my_abstract_attribute(self):
                pass

        class MyClassOK(MyABC):
            def __init__(self):
                self.my_abstract_attribute = 1

        class MyClassNotOK(MyABC):
            pass

    Raises
    ------
    NotImplementedError
        When it's not possible to instantiate an abstract class with abstract
        attributes.
    """

    def __call__(cls, *args, **kwargs):
        # Collect all abstract-attribute names from the entire MRO
        abstract_attr_names = set()
        for base in cls.__mro__:
            for name, value in base.__dict__.items():
                if getattr(value, "__is_abstract_attribute__", False):
                    abstract_attr_names.add(name)

        for name, value in cls.__dict__.items():
            if not getattr(value, "__is_abstract_attribute__", False):
                abstract_attr_names.discard(name)

        if abstract_attr_names:
            raise NotImplementedError(
                "Can't instantiate abstract class "
                "{} with abstract attributes: {}".format(
                    cls.__name__, ", ".join(abstract_attr_names)
                )
            )

        return super().__call__(*args, **kwargs)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2013-2016 Colin Duquesnoy and others (see pyqode/AUTHORS.rst)
# Copyright (c) 2016- Spyder Project Contributors (see AUTHORS.txt)
#
# Distributed under the terms of the MIT License
# (see NOTICE.txt in the Spyder root directory for details)
# -----------------------------------------------------------------------------

"""
This module contains the editor extensions controller.

Adapted from pyqode/core/managers/modes.py of the
`PyQode project <https://github.com/pyQode/pyQode>`_.
Original file:
<https://github.com/pyQode/pyqode.core/raw/master/pyqode/core/managers/modes.py>
"""

# Stdlib imports
import logging

# Local imports
from spyder.plugins.editor.api.manager import Manager


logger = logging.getLogger(__name__)


class EditorExtensionsManager(Manager):
    """Manages the list of editor extensions of the CodeEdit widget."""

    def __init__(self, editor):
        """Initialize and add a reference to the editor."""
        super().__init__(editor)
        self._extensions = {}

    def add(self, extension):
        """
        Add a extension to the editor.

        :param extension: The extension instance to add.

        """
        logger.debug('adding extension {}'.format(extension.name))
        self._extensions[extension.name] = extension
        extension.on_install(self.editor)
        return extension

    def remove(self, name_or_klass):
        """
        Remove a extension from the editor.

        :param name_or_klass: The name (or class) of the extension to remove.
        :returns: The removed extension.
        """
        logger.debug('removing extension {}'.format(name_or_klass))
        extension = self.get(name_or_klass)
        extension.on_uninstall()
        self._extensions.pop(extension.name)
        return extension

    def clear(self):
        """
        Remove all extensions from the editor.

        All extensions are removed fromlist and deleted.
        """
        while len(self._extensions):
            key = sorted(list(self._extensions.keys()))[0]
            self.remove(key)

    def get(self, name_or_klass):
        """
        Get a extension by name (or class).

        :param name_or_klass: The name or the class of the extension to get
        :type name_or_klass: str or type
        :rtype: spyder.api.mode.EditorExtension
        """
        if not isinstance(name_or_klass, str):
            name_or_klass = name_or_klass.__name__
        return self._extensions[name_or_klass]

    def keys(self):
        """
        Return the list of the names of the installed extensions.
        """
        return self._extensions.keys()

    def values(self):
        """
        Return the list of installed extensions.
        """
        return self._extensions.values()

    def __len__(self):
        """Return the amount of installed extensions."""
        return len(list(self._extensions.values()))

    def __iter__(self):
        """
        Return the list of extensions.

        :return:
        """
        return iter([v for k, v in sorted(self._extensions.items())])



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Spyder Editor-specific run schemas."""

# Standard library imports
from __future__ import annotations

from typing import Tuple, List, TypedDict

# Local imports
from spyder.plugins.run.api import Context


class EditorRunConfiguration(TypedDict):
    """Editor supported run configuration schema."""

    # Name of the plugin that is extending the editor supported run
    # configurations
    origin: str

    # Filename extension for which the editor should be able to provide
    # run configurations.
    extension: str

    # List of contexts for which the editor should produce run configurations.
    contexts: List[Context]


class FileRun(TypedDict):
    """Schema emitted by the editor for the `File` run context."""

    # File path to the file to execute. It is responsibility of the
    # executor to load and correctly interpret the file.
    path: str


class SelectionRun(TypedDict):
    """Schema emitted by the editor for the `Selection` run context."""

    # File path to the file that contains the selection to execute.
    path: str

    # Actual selection text to execute.
    selection: str

    # Encoding of the text.
    encoding: str

    # Selection start and end in (line, column) format
    line_col_bounds: Tuple[Tuple[int, int], Tuple[int, int]]

    # Selection start and end in characters
    character_bounds: Tuple[int, int]


class CellRun(TypedDict):
    """Schema emitted by the editor for the `Cell` run context."""

    # File path to the file that contains the selection to execute.
    path: str

    # Actual cell text to execute.
    cell: str

    # Name of the cell.
    cell_name: str

    # Encoding of the text.
    encoding: str

    # Selection start and end in (line, column) format
    line_col_bounds: Tuple[Tuple[int, int], Tuple[int, int]]

    # Selection start and end in characters
    character_bounds: Tuple[int, int]

    # True if the text should be copied over.
    copy: bool


class SelectionContextModificator:
    ToLine = "up to line"
    FromLine = "from line"


class ExtraAction:
    Advance = "advance"



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
spyder.plugins.remoteclient
===========================

Remote Client Plugin.
"""

# Required version of spyder-remote-services
SPYDER_REMOTE_MIN_VERSION = "1.0.0"
SPYDER_REMOTE_MAX_VERSION = "2.0.0"
SPYDER_REMOTE_VERSION = (
    f">={SPYDER_REMOTE_MIN_VERSION},<{SPYDER_REMOTE_MAX_VERSION}"
)

# jupyter server's extension name for spyder-remote-services
SPYDER_PLUGIN_NAME = "spyder-services"



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Pyenv utilities
"""

import os
import os.path as osp

from spyder.config.base import get_home_dir
from spyder.utils.programs import find_program, run_shell_command


PYENV_ENV_LIST_CACHE = {}


def get_pyenv_path(name):
    """Return the complete path of the pyenv."""
    home = get_home_dir()
    if os.name == 'nt':
        path = osp.join(
            home, '.pyenv', 'pyenv-win', 'versions', name, 'python.exe')
    elif name == '':
        path = osp.join(home, '.pyenv', 'shims', 'python')
    else:
        path = osp.join(home, '.pyenv', 'versions', name, 'bin', 'python')
    return path


def get_list_pyenv_envs():
    """Return the list of all pyenv envs found in the system."""
    global PYENV_ENV_LIST_CACHE

    env_list = {}
    pyenv = find_program('pyenv')
    if pyenv is None:
        return env_list

    cmdstr = ' '.join([pyenv, 'versions', '--bare', '--skip-aliases'])
    try:
        out, __ = run_shell_command(cmdstr, env={}).communicate()
        out = out.decode().strip()
    except Exception:
        return env_list

    out = out.split('\n') if out else []
    for env in out:
        data = env.split(osp.sep)
        path = get_pyenv_path(data[-1])

        name = f'Pyenv: {data[-1]}'
        version = f'Python {data[0]}'

        if name in env_list:
            if not (path, version) == env_list[name]:
                prev_info = env_list[name]
                prev_data = prev_info[0]
                prev_data = prev_data.split(osp.sep)
                env_list.pop(name)
                index_common_folder = 1
                for i in range(-1, -len(data) - 1, -1):
                    if data[i] == prev_data[i]:
                        index_common_folder += 1
                    else:
                        break
                part_path = prev_data[-index_common_folder : -1]
                prev_name = f'Pyenv: {"/".join(part_path)}'
                env_list[prev_name] = prev_info
                name = f'Pyenv: {"/".join(data[-index_common_folder:])}'

        
        env_list[name] = (path, version)

    PYENV_ENV_LIST_CACHE = env_list
    return env_list


def get_list_pyenv_envs_cache():
    """Return a cache of envs to avoid computing them again."""
    return PYENV_ENV_LIST_CACHE



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Module checking Spyder installation requirements"""

# Third-party imports
from packaging.version import parse


def show_warning(message):
    """Show warning using Tkinter if available"""
    try:
        # If tkinter is installed (highly probable), show an error pop-up.
        # From https://stackoverflow.com/a/17280890/438386
        import tkinter as tk
        root = tk.Tk()
        root.title("Spyder")
        label = tk.Label(root, text=message, justify='left')
        label.pack(side="top", fill="both", expand=True, padx=20, pady=20)
        button = tk.Button(root, text="OK", command=root.destroy)
        button.pack(side="bottom", fill="none", expand=True)
        root.mainloop()
    except Exception:
        pass

    raise RuntimeError(message)


def check_qt():
    """Check Qt binding requirements"""
    qt_infos = dict(
        pyqt5=("PyQt5", ("5.15.0", "5.16")),
        pyside2=("PySide2", ("5.15.0", "5.16")),
        pyqt6=("PyQt6", ("6.9.0", "7.0.0")),
        pyside6=("PySide6", ("6.8.0", "6.9.0")),
    )

    try:
        import qtpy
        package_name, required_ver = qt_infos[qtpy.API]
        actual_ver = qtpy.QT_VERSION

        if actual_ver is None or not (
            parse(required_ver[0])
            <= parse(actual_ver)
            < parse(required_ver[1])
        ):
            show_warning(
                (
                    "Please check Spyder installation requirements:\n\n"
                    "{} >={},<{} is required but version {} was found."
                ).format(
                    package_name, required_ver[0], required_ver[1], actual_ver
                )
            )
    except Exception as e:
        show_warning("Failed to import qtpy.\n\nThe error was {}".format(e))



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# Standard library imports
from __future__ import annotations  # noqa; required for typing in Python 3.8
from glob import glob
from hashlib import sha256
import logging
import os
import os.path as osp
import platform
import shutil
import subprocess
import sys
from time import sleep
import traceback
from typing import TypedDict
from zipfile import ZipFile

# Third party imports
from packaging.version import parse, Version
from qtpy.QtCore import QObject, Signal
import requests
from requests.exceptions import ConnectionError, HTTPError, SSLError
from spyder_kernels.utils.pythonenv import is_conda_env

# Local imports
from spyder import __version__
from spyder.api.translations import _
from spyder.config.base import (
    is_conda_based_app,
    is_installed_all_users,
    running_in_ci,
)
from spyder.plugins.updatemanager.utils import get_updater_info
from spyder.utils.conda import get_spyder_conda_channel, find_conda
from spyder.utils.programs import get_temp_dir, find_program

# Logger setup
logger = logging.getLogger(__name__)

CURRENT_VERSION = parse(__version__)

CONNECT_ERROR_MSG = _(
    "Unable to connect to the Spyder update service."
    "<br><br>Make sure your connection is working properly."
)

HTTP_ERROR_MSG = _(
    "HTTP error {status_code} when checking for updates."
    "<br><br>Make sure your connection is working properly,"
    "and try again later."
)

SSL_ERROR_MSG = _(
    "SSL certificate verification failed while checking for Spyder updates."
    "<br><br>Please contact your network administrator for assistance."
)

OS_ERROR_MSG = _(
    "An error occurred while checking for Spyder updates, possibly related to "
    "your operating system configuration or file access.<br><br>If you're not "
    "sure what to do about it, you can disable checking for updates below. "
    "<br><br>The error was:<br><br><i>{error}</i>"
)

GH_HEADERS = {"Accept": "application/vnd.github+json"}
_token = os.getenv("GITHUB_TOKEN")
if running_in_ci() and _token:
    GH_HEADERS.update(Authorization=f"Bearer {_token}")


class UpdateType:
    """Enum with the different update types."""

    Major = "major"
    Minor = "minor"
    Micro = "micro"


class AssetInfo(TypedDict):
    """Schema for asset information."""

    # Version
    version: Version

    # Filename with extension of the release asset to download.
    filename: str

    # Type of update
    update_type: UpdateType

    # Download URL for the asset.
    url: str

    # File sha256 checksum
    checksum: str


def get_github_releases(
    tags: str | tuple[str] | None = None, updater: bool = False
) -> dict[Version, dict]:
    """
    Get Github release information

    Parameters
    ----------
    tags : str | tuple[str] | (None)
        If tags is provided, only release information for the requested tags
        is retrieved. Otherwise, the most recent 20 releases are retrieved.
        This is only used to retrieve a known set of releases for unit testing.
    updater : bool (False)
        Whether to get Spyder-updater releases (True) or Spyder releases
        (False).

    Returns
    -------
    releases : dict[packaging.version.Version, dict]
        Dictionary of release information.
    """
    url = "https://api.github.com/repos/{}/releases".format(
        "spyder-ide/spyder-updater" if updater else "spyder-ide/spyder"
    )

    if tags is None:
        # Get 20 most recent releases
        url += "?per_page=20&page=1"
        logger.info(f"Getting release info from {url}")
        page = requests.get(url, headers=GH_HEADERS)
        page.raise_for_status()
        data = page.json()
    else:
        # Get specified releases
        tags = [tags] if isinstance(tags, str) else tags
        url += "/tags/"

        data = []
        with requests.Session() as session:
            session.headers = GH_HEADERS
            logger.info(f"Getting release info for {tags}")
            for tag in tags:
                _url = url + tag
                page = session.get(_url)
                page.raise_for_status()
                data.append(page.json())

    return {parse(item["tag_name"]): item for item in data}


def get_asset_info(
    release_info: dict, current_version: Version, updater: bool
) -> AssetInfo | None:
    """
    Get the version, name, update type, download URL, and digest for the asset
    of the given release.

    Parameters
    ----------
    release_info: dict
        Release information from Github for a single release.
    current_version: Version
        Current version.
    updater: bool
        Whether Spyder-updater (True) or Spyder (False).

    Returns
    -------
    asset_info: AssetInfo | None
        Information about the asset.
    """
    release = parse(release_info["tag_name"])

    if current_version.major < release.major or not is_conda_based_app():
        update_type = UpdateType.Major
    elif (
        current_version.minor < release.minor or current_version.is_prerelease
    ):
        update_type = UpdateType.Minor
    else:
        update_type = UpdateType.Micro

    if updater:
        filename = "spyder-updater.zip"
    elif update_type != UpdateType.Major:
        filename = "spyder-conda-lock.zip"
    else:
        machine = platform.machine().lower().replace("amd64", "x86_64")
        if os.name == "nt":
            filename = f"Spyder-Windows-{machine}.exe"
        if sys.platform == "darwin":
            filename = f"Spyder-macOS-{machine}.pkg"
        if sys.platform.startswith("linux"):
            filename = f"Spyder-Linux-{machine}.sh"

    asset_info = None
    for asset in release_info["assets"]:
        if asset["name"] == filename:
            asset_info = AssetInfo(
                version=release,
                update_type=update_type,
                filename=asset["name"],
                url=asset["browser_download_url"],
                checksum=asset["digest"],
            )
            break

    return asset_info


def _check_asset_available(
    releases: dict[Version, dict],
    current_version: Version,
    updater: bool = False,
) -> AssetInfo | None:
    latest_release = max(releases) if releases else current_version
    update_available = current_version < latest_release
    asset_info = None

    logger.debug(f"Latest release: {latest_release}")
    logger.debug(f"Update available: {update_available}")

    # Check if the asset is available for download.
    # If the asset is not available, then check the next latest
    # release, and so on until either a new asset is available or there
    # is no update available.
    while update_available:
        asset_info = get_asset_info(
            releases.get(latest_release), current_version, updater
        )

        if asset_info is not None:
            break

        # The asset is not available
        logger.debug(f"Asset not available: {latest_release}")
        releases.pop(latest_release)

        latest_release = max(releases) if releases else current_version
        update_available = current_version < latest_release

        logger.debug(f"Latest release: {latest_release}")
        logger.debug(f"Update available: {update_available}")

    return asset_info


def validate_download(file: str, checksum: str) -> bool:
    """
    Compute the sha256 checksum of the provided file and compare against
    the provided checksum.

    Parameters
    ----------
    file : str
        Full path to the file to be verified.
    checksum : str
        sha256 checksum to match against the computed checksum.

    Returns
    -------
    valid : bool
        True if the file's computed checksum matches the provided checksum,
        False otherwise.
    """
    with open(file, "rb") as f:
        _checksum = sha256()
        while chunk := f.read(8192):
            _checksum.update(chunk)

    valid = checksum.endswith(_checksum.hexdigest())
    logger.debug(f"Valid {file}: {valid}")

    # Extract validated zip files
    if valid and file.endswith(".zip"):
        with ZipFile(file, "r") as f:
            f.extractall(osp.dirname(file))
        logger.debug(f"{file} extracted.")

    return valid


class UpdateDownloadCancelledException(Exception):
    """Download for installer to update was cancelled."""

    pass


class UpdateDownloadError(Exception):
    """Error occured while downloading file"""

    pass


class UpdateUpdaterUACCancelled(Exception):
    """UAC elevation was cancelled"""

    pass


class BaseWorker(QObject):
    """Base worker class for the updater"""

    sig_ready = Signal(bool)
    """
    Signal to inform that the worker has finished.

    Parameters
    ----------
    success : bool
        Whether the worker was successful (True) or not (False).
    """

    sig_exception_occurred = Signal(dict)
    """
    Send untracked exceptions to the error reporter

    Parameters
    ----------
    error_data: dict
        The dictionary containing error data. The allowed keys are:
        text: str
            Error text to display. This may be a translated string or
            formatted exception string.
        is_traceback: bool
            Whether `text` is plain text or an error traceback.
        repo: str
            Customized display of repo in GitHub error submission report.
        title: str
            Customized display of title in GitHub error submission report.
        label: str
            Customized content of the error dialog.
        steps: str
            Customized content of the error dialog.
    """


class WorkerUpdate(BaseWorker):
    """
    Worker that checks for releases using either the Anaconda
    default channels or the Github Releases page without
    blocking the Spyder user interface, in case of connection
    issues.
    """

    def __init__(self, stable_only):
        super().__init__()
        self.stable_only = stable_only
        self.asset_info = None
        self.error = None
        self.checkbox = False
        self.channel = None

    def _check_update_available(self, release: Version | None = None):
        """
        Check if there is an update available.

        Releases are obtained from Github and compared to the current Spyder
        version to determine if an update is available. The Github release
        asset and checksum must be available in order for the release to be
        considered.

        Parameters
        ----------
        release : packaging.version.Version | (None)
            If provided, limit possible releases on Github to this release
            or less.
        """

        # Get asset info from Github
        releases = get_github_releases()

        if release is not None:
            # Limit Github releases to consider
            releases = {k: v for k, v in releases.items() if k <= release}

        if self.stable_only:
            # Only consider stable releases
            releases = {
                k: v for k, v in releases.items() if not k.is_prerelease
            }
        logger.debug(f"Available releases: {sorted(releases)}")

        self.asset_info = _check_asset_available(releases, CURRENT_VERSION)

    def start(self):
        """Main method of the worker."""
        url = None
        if not is_conda_based_app():
            self.channel = "pypi"  # Default channel if not conda
            if is_conda_env(sys.prefix):
                self.channel, url = get_spyder_conda_channel()

            if self.channel == "pypi":
                url = "https://pypi.python.org/pypi/spyder/json"
            elif url is not None:
                url += "/channeldata.json"

        try:
            release = None
            if url is not None:
                # Limit the releases on Github that we consider to those less
                # than or equal to what is also available on the conda/pypi
                # channel
                logger.info(f"Getting release from {url}")
                page = requests.get(url)
                page.raise_for_status()
                data = page.json()

                if self.channel == "pypi":
                    releases = [
                        parse(k)
                        for k in data["releases"].keys()
                        if not parse(k).is_prerelease
                    ]
                    release = max(releases)
                else:
                    # Conda pkgs/main or conda-forge url
                    spyder_data = data["packages"].get("spyder")
                    if spyder_data:
                        release = parse(spyder_data["version"])

            if (
                # Always check for updates for our installers or pip
                # installations
                is_conda_based_app()
                or self.channel == "pypi"
                # Only offer updates for conda envs if we're able to detect the
                # channel. That's because Anaconda is always several versions
                # behind the latest one.
                or is_conda_env(sys.prefix) and url is not None
            ):
                self._check_update_available(release)

        except SSLError as err:
            self.error = SSL_ERROR_MSG
            logger.warning(err, exc_info=err)
        except ConnectionError as err:
            self.error = CONNECT_ERROR_MSG
            logger.warning(err, exc_info=err)
        except HTTPError as err:
            status_code = err.response.status_code
            self.error = HTTP_ERROR_MSG.format(status_code=status_code)
            logger.warning(err, exc_info=err)
        except OSError as err:
            self.error = OS_ERROR_MSG.format(error=err)
            self.checkbox = True
            logger.warning(err, exc_info=err)
        except Exception as err:
            # Send untracked errors to our error reporter
            error_data = dict(
                text=traceback.format_exc(),
                is_traceback=True,
                title="Error when checking for updates",
            )
            self.sig_exception_occurred.emit(error_data)
            logger.error(err, exc_info=err)
        finally:
            # At this point we **must** emit the signal below so that the
            # "Check for updates" action in the Help menu is enabled again
            # after the check has finished (it's disabled while the check is
            # running).
            try:
                self.sig_ready.emit(self.error is None)
            except RuntimeError:
                pass


class WorkerUpdateUpdater(BaseWorker):
    """
    Worker that checks and updates Spyder-updater without blocking
    the Spyder user interface.
    """

    def __init__(self, stable_only):
        super().__init__()
        self.stable_only = stable_only
        self.asset_info = None
        self.installer_path = None
        self.error = None
        __, updater_version = get_updater_info()
        self.updater_version = parse(updater_version)

    def _check_asset_available(self):
        """Checks if there is an update available for the Updater."""

        # Get release info from Github
        releases = get_github_releases(updater=True)

        if self.stable_only:
            # Only use stable releases
            releases = {
                k: v for k, v in releases.items() if not k.is_prerelease
            }
        logger.debug(f"Available releases: {sorted(releases)}")

        self.asset_info = _check_asset_available(
            releases, self.updater_version, updater=True
        )

    def _clean_installer_dir(self):
        """Remove downloaded file"""
        installer_dir = osp.dirname(self.installer_path)
        if osp.exists(installer_dir):
            try:
                shutil.rmtree(installer_dir)
            except OSError as err:
                logger.debug(err, stack_info=True)

    def _download_asset(self):
        """Download Updater lock file"""
        self.installer_path = osp.join(
            get_temp_dir(),
            "updates",
            "spyder-updater",
            str(self.asset_info["version"]),
            self.asset_info["filename"],
        )

        if osp.exists(self.installer_path) and validate_download(
            self.installer_path, self.asset_info["checksum"]
        ):
            logger.debug(f"{self.installer_path} already downloaded.")
            return

        self._clean_installer_dir()
        dirname = osp.dirname(self.installer_path)
        os.makedirs(dirname, exist_ok=True)

        url = self.asset_info["url"]
        logger.info(f"Downloading {url} to {self.installer_path}")

        page = requests.get(url, headers=GH_HEADERS)
        page.raise_for_status()
        with open(self.installer_path, "wb") as f:
            f.write(page.content)

        if validate_download(self.installer_path, self.asset_info["checksum"]):
            logger.info("Download successfully completed.")
        else:
            raise UpdateDownloadError("Download failed!")

    def _install_update(self):
        """
        Create or update Spyder-updater environment.

        In the case where Spyder is installed for all users on Windows,
        User Account Control (UAC) needs to be elevated in order to update
        Spyder-updater. This requires that the subprocess run through
        PowerShell with the `-Verb RunAs` flag. However, this flag is
        incompatible with capturing std[out|err], either with a subprocess.PIPE
        or with `-RedirectStandard[Output|Error]`. Thus, the only solution is
        to run a separate script that internally redirects std[out|err] to
        (temporary) log files for later inspection.

        Unix systems do not require a separate script, but it is used to
        retain simplicity and parity between the platforms.

        Note that ctypes.windll.shell32.ShellExecuteEx is inadequate since it
        does not wait for the process to complete before returning.
        """
        updater_script = osp.join(
            osp.dirname(__file__),
            "scripts",
            "updater.bat" if os.name == "nt" else "updater.sh",
        )
        conda_exe = find_conda()
        conda_cmd = "create"
        if self.updater_version > parse("0.0.0"):
            conda_cmd = "update"
        env_path = osp.join(osp.dirname(sys.prefix), "spyder-updater")

        installer_dir = osp.dirname(self.installer_path)
        if os.name == "nt":
            plat = "win-64"
        elif sys.platform == "darwin":
            plat = "osx-arm64" if platform.machine() == "arm64" else "osx-64"
        else:
            plat = "linux-64"
        spy_updater_lock = osp.join(
            installer_dir, f"conda-updater-{plat}.lock"
        )
        spy_updater_conda = glob(
            osp.join(installer_dir, "spyder-updater*.conda")
        )[0]

        # Run updater script
        kwargs = dict(shell=True, capture_output=True, text=True)
        cmd = [
            updater_script,
            conda_exe,
            conda_cmd,
            env_path,
            spy_updater_lock,
            spy_updater_conda,
        ]
        if os.name == "nt":
            kwargs.update(executable=find_program("powershell"))
            cmd = [
                "start",
                "-FilePath",
                f'"{updater_script}"',
                "-ArgumentList",
                ",".join([f"'{a}'" for a in cmd[1:]]),
                "-Wait",
                "-WindowStyle",
                "Hidden",
            ]
            if is_installed_all_users():
                cmd.extend(["-Verb", "RunAs"])

        logger.info("Updating Spyder Updater...")
        proc = subprocess.run(" ".join(cmd), **kwargs)

        # Check for errors
        if os.name == "nt":
            # Stdout/err is only captured for the high-level subprocess.
            # If there is an error here, it will only be if the user
            # cancelled the UAC elevation
            try:
                proc.check_returncode()
            except subprocess.CalledProcessError as err:
                raise UpdateUpdaterUACCancelled(err.stderr)

            # Stdout/err is not captured by subprocess for the updater_script
            # Check for errors there by reading the logs
            updater_stdout = osp.join(installer_dir, "updater_stdout.log")
            updater_stderr = osp.join(installer_dir, "updater_stderr.log")
            with open(updater_stderr, "r") as f:
                stderr = f.read()
            with open(updater_stdout, "r") as f:
                stdout = f.read()
            if stderr:
                raise subprocess.CalledProcessError(
                    1, " ".join(cmd), output=stdout, stderr=stderr
                )
        else:
            proc.check_returncode()

    def start(self):
        """Main method of the worker."""

        try:
            self._check_asset_available()
            if (
                self.asset_info is None
                and self.updater_version == parse("0.0.0")
            ):
                raise RuntimeError(
                    "Spyder-updater is not installed and "
                    "not available for download!"
                )
            elif self.asset_info is not None:
                self._download_asset()
                self._install_update()
        except UpdateUpdaterUACCancelled as err:
            self.error = err
            logger.debug(err)
        except subprocess.CalledProcessError as err:
            self.error = err
            logger.debug(err)
        except Exception as err:
            # Send untracked errors to our error reporter
            self.error = str(err)
            error_data = dict(
                text=traceback.format_exc(),
                is_traceback=True,
                title=_("Error when updating Spyder-updater"),
            )
            self.sig_exception_occurred.emit(error_data)
            logger.error(err, exc_info=err)
        finally:
            # At this point we **must** emit the signal below so that the
            # "Check for updates" action in the Help menu is enabled again
            # after the check has finished (it's disabled while the check is
            # running).
            try:
                self.sig_ready.emit(self.error is None)
            except RuntimeError:
                pass


class WorkerDownloadInstaller(BaseWorker):
    """
    Worker that donwloads standalone installers for Windows, macOS,
    and Linux without blocking the Spyder user interface.
    """

    sig_download_progress = Signal(int, int)
    """
    Signal to send the download progress.

    Parameters
    ----------
    current_value: int
        Size of the data downloaded until now.
    total: int
        Total size of the file expected to be downloaded.
    """

    def __init__(self, asset_info, installer_path):
        super().__init__()
        self.asset_info = asset_info
        self.installer_path = installer_path
        self.error = None
        self.cancelled = False
        self.paused = False

    def _progress_reporter(self, progress, total_size):
        """Calculate download progress and notify."""
        self.sig_download_progress.emit(progress, total_size)

        while self.paused and not self.cancelled:
            sleep(0.5)

        if self.cancelled:
            raise UpdateDownloadCancelledException()

    def _download_installer(self):
        """Donwload Spyder installer."""
        url = self.asset_info["url"]
        logger.info(f"Downloading {url} to {self.installer_path}")

        self._clean_installer_dir()
        installer_dir = osp.dirname(self.installer_path)
        os.makedirs(installer_dir, exist_ok=True)

        with requests.get(url, stream=True) as r:
            r.raise_for_status()

            size = int(r.headers["content-length"])
            self._progress_reporter(0, size)

            with open(self.installer_path, "wb") as f:
                chunk_size = 8 * 1024
                size_read = 0
                for chunk in r.iter_content(chunk_size=chunk_size):
                    size_read += len(chunk)
                    f.write(chunk)
                    self._progress_reporter(size_read, size)

        if validate_download(self.installer_path, self.asset_info["checksum"]):
            logger.info("Download successfully completed.")
        else:
            raise UpdateDownloadError("Download failed!")

    def _clean_installer_dir(self):
        """Remove downloaded file"""
        installer_dir = osp.dirname(self.installer_path)
        if osp.exists(installer_dir):
            try:
                shutil.rmtree(installer_dir)
            except OSError as err:
                logger.debug(err, stack_info=True)

    def start(self):
        """Main method of the worker."""
        try:
            self._download_installer()
        except UpdateDownloadCancelledException:
            logger.info("Download cancelled")
            self._clean_installer_dir()
        except SSLError as err:
            self.error = SSL_ERROR_MSG
            logger.warning(err, exc_info=err)
        except ConnectionError as err:
            self.error = CONNECT_ERROR_MSG
            logger.warning(err, exc_info=err)
        except Exception as err:
            error = traceback.format_exc()
            formatted_error = error.replace("\n", "<br>").replace(
                " ", "&nbsp;"
            )

            self.error = _(
                "It was not possible to download the installer due to the "
                "following error:"
                "<br><br>"
                "<tt>{}</tt>"
            ).format(formatted_error)
            logger.warning(err, exc_info=err)
            self._clean_installer_dir()
        finally:
            try:
                self.sig_ready.emit(self.error is None)
            except RuntimeError:
                pass
